import multiprocessing as mp

import pytest
# noinspection PyProtectedMember
from dynesty.dynesty import _function_wrapper
from emcee.ensemble import _FunctionWrapper

import autofit as af
from autoconf.conf import output_path_for_test
from autofit.non_linear.parallel import SneakyPool, SneakyJob
# noinspection PyProtectedMember
from autofit.non_linear.parallel.sneaky import _is_likelihood_function


@pytest.fixture(
    name="search"
)
def make_search():
    return af.DynestyStatic(
        name="",
        number_of_cores=2,
        maxcall=2
    )


@pytest.fixture(
    name="model"
)
def make_model():
    return af.Model(
        af.Gaussian
    )


@pytest.fixture(
    name="analysis"
)
def make_analysis():
    return Analysis()


@pytest.fixture(
    name="paths"
)
def make_paths(
        model,
        search
):
    paths = search.paths
    paths.model = model
    return model


@pytest.fixture(
    name="fitness"
)
def make_fitness(
        paths,
        model,
        analysis,
):
    return Fitness(
        paths=paths,
        model=model,
        analysis=analysis,
        samples_from_model=None
    )


class Analysis(af.Analysis):
    def log_likelihood_function(self, instance: af.Gaussian):
        return -(10 - instance.centre) ** 2


class Fitness(af.NonLinearSearch.Fitness):
    def figure_of_merit_from(self, parameter_list):
        return -1

    def __call__(self, parameter_list):
        return self.figure_of_merit_from(
            parameter_list
        )


@pytest.fixture(
    name="sneaky_job"
)
def make_sneaky_job(
        fitness
):
    return SneakyJob(
        identity,
        1,
        fitness
    )


def test_perform_fitness(
        pool,
        fitness
):
    results = list(pool.map(
        fitness,
        range(2)
    ))
    assert results == [-1, -1]


def test_sneaky_removal(
        sneaky_job
):
    assert sneaky_job.args == [1]
    assert sneaky_job.fitness_index == 1


def test_sneaky_call(
        sneaky_job,
        fitness
):
    assert sneaky_job.perform(
        fitness
    ) == ([1, fitness],)


def identity(*args):
    return args


def test_sneaky_pool(
        fitness,
        pool
):
    results = list(pool.map(identity, [(0, fitness), (1, fitness)]))

    assert len(results) == 2

    for i, t in enumerate(results):
        assert t[0][0] == i
        assert isinstance(
            t[0][1],
            Fitness
        )


def test_arbitrary_map(
        pool
):
    args_list = [(1,), (1, 2), (1, 2, 3)]
    results = list(pool.map(
        identity,
        args_list
    ))

    assert results == [([1],), ([1, 2],), ([1, 2, 3],)]


def get_pid(*_):
    return mp.current_process().pid


@pytest.fixture(
    name="pool"
)
def make_pool(fitness):
    return SneakyPool(
        processes=1,
        paths=af.DirectoryPaths(name="test"),
        fitness=fitness
    )


def test_process_ids(pool):
    results = set(pool.map(
        get_pid,
        range(10)
    ))
    assert results == set(pool.map(
        get_pid,
        range(10)
    ))


class MyException(Exception):
    pass


def raise_exception(*_):
    raise MyException()


def test_raising_error(pool):
    with pytest.raises(
            MyException
    ):
        list(pool.map(
            raise_exception,
            range(10)
        ))


@output_path_for_test()
def test_sneaky_map(
        search,
        model,
        analysis
):
    result = search.fit(
        model,
        analysis
    )

    assert isinstance(
        result.instance,
        af.Gaussian
    )


@pytest.mark.parametrize(
    "function",
    [
        _FunctionWrapper(lambda: 1, None, None),
        Fitness(None, None, None, None),
        _function_wrapper(lambda: 1, [], {}, "loglikelihood")
    ]
)
def test_is_likelihood_function(
        function
):
    assert _is_likelihood_function(
        function
    )



import numpy as np

from autoconf import conf
from autofit import exc
from autofit.messages.normal import NormalMessage, UniformNormalMessage
from autofit.messages.transform import log_10_transform
from autofit.messages.transform_wrapper import TransformedWrapperInstance
from .abstract import epsilon, assert_within_limits


class Limits:
    @staticmethod
    def for_class_and_attributes_name(cls, attribute_name):
        limit_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name, "gaussian_limits"]
        )
        return limit_dict["lower"], limit_dict["upper"]


class WrappedInstance(
    TransformedWrapperInstance
):
    __identifier_fields__ = ("lower_limit", "upper_limit")

    __database_args__ = (
        "lower_limit",
        "upper_limit",
        "log_norm",
        "id_",
    )

    def __init__(
            self,
            transformed_wrapper,
            *args,
            lower_limit,
            upper_limit,
            **kwargs
    ):
        super().__init__(
            transformed_wrapper,
            *args,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            **kwargs
        )
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

        if self.lower_limit >= self.upper_limit:
            raise exc.PriorException(
                "The upper limit of a prior must be greater than its lower limit"
            )

    def _new_for_base_message(
            self,
            message
    ):
        return type(self)(
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
            id_=self.instance().id,
            params=message.parameters
        )


class UniformPrior(WrappedInstance):
    """A prior with a uniform distribution between a lower and upper limit"""

    def __init__(
            self,
            lower_limit=0.0,
            upper_limit=1.0,
            id_=None,
            params=(0.0, 1.0)
    ):
        if any(map(np.isnan, params)):
            raise exc.MessageException(
                "nan parameter passed to UniformPrior"
            )
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)

        Message = UniformNormalMessage.shifted(
            shift=lower_limit,
            scale=upper_limit - lower_limit,
        )
        super().__init__(
            Message,
            *params,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            id_=id_
        )

    def logpdf(self, x):
        # TODO: handle x as a numpy array
        if x == self.lower_limit:
            x += epsilon
        elif x == self.upper_limit:
            x -= epsilon
        return self.instance().logpdf(x)

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"UniformPrior, lower_limit = {self.lower_limit}, upper_limit = {self.upper_limit}"

    @assert_within_limits
    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute between the upper and lower limits
        """
        return round(super().value_for(unit), 14)

    # noinspection PyUnusedLocal
    @staticmethod
    def log_prior_from_value(value):
        """
        Returns the log prior of a physical value, so the log likelihood of a model evaluation can be converted to a
        posterior as log_prior + log_likelihood.

        This is used by Emcee in the log likelihood function evaluation.

        NOTE: For a UniformPrior this is always zero, provided the value is between the lower and upper limit. Given
        this is check for when the instance is made (in the *instance_from_vector* function), we thus can simply return
        zero in this function.
        """
        return 0.0


class LogUniformPrior(WrappedInstance):
    """A prior with a uniform distribution between a lower and upper limit"""

    def __init__(
            cls,
            lower_limit=1e-6,
            upper_limit=1.0,
            id_=None,
            params=(0.0, 1.0)
    ):
        if lower_limit <= 0.0:
            raise exc.PriorException(
                "The lower limit of a LogUniformPrior cannot be zero or negative."
            )

        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)

        Message = UniformNormalMessage.shifted(
            shift=np.log10(lower_limit),
            scale=np.log10(upper_limit / lower_limit),
        ).transformed(
            log_10_transform
        )

        super().__init__(
            Message,
            *params,
            id_=id_,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

    __identifier_fields__ = ("lower_limit", "upper_limit")

    @staticmethod
    def log_prior_from_value(value):
        """
        Returns the log prior of a physical value, so the log likelihood of a model evaluation can be converted to a
            posterior as log_prior + log_likelihood.

        This is used by Emcee in the log likelihood function evaluation.

        Parameters
        ----------
        value : float
            The physical value of this prior's corresponding parameter in a `NonLinearSearch` sample."""
        return 1.0 / value

    @assert_within_limits
    def value_for(self, unit: float) -> float:
        return super().value_for(unit)

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"LogUniformPrior, lower_limit = {self.lower_limit}, upper_limit = {self.upper_limit}"


class GaussianPrior(NormalMessage):
    """A prior with a gaussian distribution"""

    __identifier_fields__ = (
        "lower_limit",
        "upper_limit",
        "mean",
        "sigma"
    )

    @assert_within_limits
    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1

        Returns
        -------
        value: Float
            A value for the attribute biased to the gaussian distribution
        """
        return super().value_for(unit)



import copy

import pytest

import autofit as af
from autofit.mock.mock_model import (
    MockClassx2,
    MockComplexClass,
    MockComponents,
    MockListClass,
    MockChildTuplex2,
    MockChildTuplex3
)

@pytest.fixture(name="instance_prior_model")
def make_instance_prior_model():
    instance = MockClassx2(1.0, 2.0)
    return af.AbstractPriorModel.from_instance(instance)


@pytest.fixture(name="list_prior_model")
def make_list_prior_model():
    instance = [MockClassx2(1.0, 2.0)]
    return af.AbstractPriorModel.from_instance(instance)


@pytest.fixture(name="complex_prior_model")
def make_complex_prior_model():
    instance = MockComplexClass(MockClassx2(1.0, 2.0))
    return af.AbstractPriorModel.from_instance(instance)


def test_class_assertion():
    with pytest.raises(AssertionError):
        af.Model("Hello")


class TestAsModel:
    def test_instance(self, instance_prior_model):
        model = instance_prior_model.as_model()
        assert model.prior_count == 2

    def test_complex(self, complex_prior_model):
        assert complex_prior_model.prior_count == 0
        model = complex_prior_model.as_model()
        assert model.prior_count == 2
        assert model.simple.prior_count == 2

    def test_galaxies(self):
        galaxies = af.ModelInstance()
        galaxies.one = MockComponents()
        instance = af.ModelInstance()
        instance.galaxies = galaxies
        model = instance.as_model(model_classes=(MockComponents,))
        assert model.prior_count == 1


class TestFromInstance:
    def test_model_mapper(self):
        instance = af.ModelInstance()
        instance.simple = MockClassx2(1.0, 2.0)

        result = af.AbstractPriorModel.from_instance(instance)

        assert isinstance(result, af.ModelMapper)

    def test_with_model_classes(self):
        instance = MockComplexClass(MockClassx2(1.0, 2.0))
        model = af.AbstractPriorModel.from_instance(
            instance, model_classes=(MockClassx2,)
        )
        assert model.prior_count == 2

    def test_list_with_model_classes(self):
        instance = [
            MockClassx2(1.0, 2.0),
            MockComplexClass(MockClassx2(1.0, 2.0)),
        ]
        model = af.AbstractPriorModel.from_instance(
            instance, model_classes=(MockComplexClass,)
        )

        assert model.prior_count == 2
        assert model[0].prior_count == 0
        assert model[1].prior_count == 2

    def test_dict_with_model_classes(self):
        instance = {
            "one": MockClassx2(1.0, 2.0),
            "two": MockComplexClass(MockClassx2(1.0, 2.0)),
        }
        model = af.AbstractPriorModel.from_instance(
            instance, model_classes=(MockComplexClass,)
        )

        assert model.prior_count == 2
        assert model[0].prior_count == 0
        assert model[1].prior_count == 2

        assert model.one.one == 1.0
        assert model.one.two == 2.0

        assert isinstance(model.two.simple.one, af.Prior)
        assert isinstance(model.two.simple.two, af.Prior)

    def test_instance(self, instance_prior_model):
        assert instance_prior_model.cls == MockClassx2
        assert instance_prior_model.prior_count == 0
        assert instance_prior_model.one == 1.0
        assert instance_prior_model.two == 2.0

        new_instance = instance_prior_model.instance_for_arguments({})
        assert isinstance(new_instance, MockClassx2)
        assert new_instance.one == 1.0
        assert new_instance.two == 2.0

    def test_complex(self, complex_prior_model):
        assert complex_prior_model.cls == MockComplexClass
        assert complex_prior_model.prior_count == 0
        assert isinstance(complex_prior_model.simple, af.PriorModel)
        assert complex_prior_model.simple.cls == MockClassx2
        assert complex_prior_model.simple.one == 1.0

        new_instance = complex_prior_model.instance_for_arguments({})
        assert isinstance(new_instance, MockComplexClass)
        assert isinstance(new_instance.simple, MockClassx2)
        assert new_instance.simple.one == 1.0

    def test_list(self, list_prior_model):
        assert isinstance(list_prior_model, af.CollectionPriorModel)
        assert isinstance(list_prior_model[0], af.PriorModel)
        assert list_prior_model[0].one == 1.0

    def test_dict(self):
        instance = {"simple": MockClassx2(1.0, 2.0)}
        prior_model = af.AbstractPriorModel.from_instance(instance)
        assert isinstance(prior_model, af.CollectionPriorModel)
        assert isinstance(prior_model.simple, af.PriorModel)
        assert prior_model.simple.one == 1.0

        new_instance = prior_model.instance_for_arguments({})
        assert isinstance(new_instance.simple, MockClassx2)

        prior_model = af.AbstractPriorModel.from_instance(new_instance)
        assert isinstance(prior_model, af.CollectionPriorModel)
        assert isinstance(prior_model.simple, af.PriorModel)
        assert prior_model.simple.one == 1.0


class TestSum:
    def test_add_prior_models(self):
        mock_cls_0 = af.PriorModel(MockChildTuplex2)
        mock_cls_1 = af.PriorModel(MockChildTuplex2)

        mock_cls_0.one = 1.0
        mock_cls_1.two = 0.0

        result = mock_cls_0 + mock_cls_1

        assert isinstance(result, af.PriorModel)
        assert result.cls == MockChildTuplex2
        assert isinstance(result.one, af.Prior)
        assert isinstance(result.two, af.Prior)

    def test_fail_for_mismatch(self):
        mock_cls_0 = af.PriorModel(MockChildTuplex2)
        mock_cls_1 = af.PriorModel(MockChildTuplex3)

        with pytest.raises(TypeError):
            mock_cls_0 + mock_cls_1

    def test_add_children(self):
        mock_components_1 = af.PriorModel(
            MockComponents,
            components_0=af.CollectionPriorModel(mock_cls_0=MockChildTuplex2),
            components_1=af.CollectionPriorModel(
                mock_cls_2=MockChildTuplex3
            ),
        )
        mock_components_2 = af.PriorModel(
            MockComponents,
            components_0=af.CollectionPriorModel(mock_cls_1=MockChildTuplex2),
            components_1=af.CollectionPriorModel(
                mock_cls_3=MockChildTuplex3
            ),
        )

        result = mock_components_1 + mock_components_2

        assert result.components_0.mock_cls_0 == mock_components_1.components_0.mock_cls_0
        assert result.components_0.mock_cls_1 == mock_components_2.components_0.mock_cls_1

        assert result.components_1.mock_cls_2 == mock_components_1.components_1.mock_cls_2
        assert result.components_1.mock_cls_3 == mock_components_2.components_1.mock_cls_3

    def test_prior_model_override(self):
        mock_components_1 = af.PriorModel(
            MockComponents,
            components_0=af.CollectionPriorModel(light=MockChildTuplex2()),
            components_1=af.CollectionPriorModel(mass=MockChildTuplex3),
        )
        mock_components_2 = af.PriorModel(
            MockComponents,
            components_0=af.CollectionPriorModel(light=MockChildTuplex2),
            components_1=af.CollectionPriorModel(
                mass=MockChildTuplex3()
            ),
        )

        result = mock_components_1 + mock_components_2

        assert result.components_1.mass == mock_components_1.components_1.mass
        assert result.components_0.light == mock_components_2.components_0.light


class TestFloatAnnotation:
    # noinspection PyUnresolvedReferences
    def test_prior_linking(self):
        mapper = af.ModelMapper()
        mapper.a = MockClassx2
        mapper.b = MockClassx2

        assert mapper.prior_count == 4

        mapper.a.one = mapper.b.one

        assert mapper.prior_count == 3

        mapper.a.two = mapper.b.two

        assert mapper.prior_count == 2

        mapper.a.one = mapper.a.two
        mapper.b.one = mapper.b.two

        assert mapper.prior_count == 1


class TestHashing:
    def test_is_hashable(self):
        assert hash(af.AbstractPriorModel()) is not None
        assert hash(af.PriorModel(MockClassx2)) is not None
        assert (
            hash(af.AnnotationPriorModel(MockClassx2, MockClassx2, "one"))
            is not None
        )

    def test_prior_prior_model_hash_consecutive(self):
        prior = af.UniformPrior(0, 1)
        prior_model = af.AbstractPriorModel()

        assert prior.id + 1 == prior_model.id


class StringDefault:
    def __init__(self, value="a string"):
        self.value = value


class TestStringArguments:
    def test_string_default(self):
        prior_model = af.PriorModel(StringDefault)
        assert prior_model.prior_count == 0

        assert prior_model.instance_for_arguments({}).value == "a string"


class TestPriorModelArguments:
    def test_list_arguments(self):
        prior_model = af.PriorModel(MockListClass)

        assert prior_model.prior_count == 0

        prior_model = af.PriorModel(MockListClass, ls=[MockClassx2])

        assert prior_model.prior_count == 2

        prior_model = af.PriorModel(
            MockListClass, ls=[MockClassx2, MockClassx2]
        )

        assert prior_model.prior_count == 4

    def test_float_argument(self):
        prior = af.UniformPrior(0.5, 2.0)
        prior_model = af.PriorModel(MockComponents, parameter=prior)

        assert prior_model.prior_count == 1
        assert prior_model.priors[0] is prior

        prior_model = af.PriorModel(MockComponents, parameter=4.0)
        assert prior_model.prior_count == 0
        assert prior_model.parameter == 4.0

        instance = prior_model.instance_for_arguments({})
        assert instance.parameter == 4.0

    def test_arbitrary_keyword_arguments(self):
        prior_model = af.PriorModel(
            MockComponents,
            mock_cls_0=MockChildTuplex2,
            mock_cls_1=MockChildTuplex3,
        )
        assert prior_model.prior_count == 10
        instance = prior_model.instance_from_unit_vector(
            [0.5] * prior_model.prior_count
        )
        assert isinstance(instance.mock_cls_0, MockChildTuplex2)
        assert isinstance(instance.mock_cls_1, MockChildTuplex3)


class TestCase:
    def test_complex_class(self):
        prior_model = af.PriorModel(MockComplexClass)

        assert hasattr(prior_model, "simple")
        assert prior_model.simple.prior_count == 2
        assert prior_model.prior_count == 2

    def test_create_instance(self):
        mapper = af.ModelMapper()
        mapper.complex = MockComplexClass

        instance = mapper.instance_from_unit_vector([1.0, 0.0])

        assert instance.complex.simple.one == 1.0
        assert instance.complex.simple.two == 0.0

    def test_instantiate_with_list_arguments(self):
        mapper = af.ModelMapper()
        mapper.list_object = af.PriorModel(
            MockListClass, ls=[MockClassx2, MockClassx2]
        )

        assert len(mapper.list_object.ls) == 2

        assert mapper.list_object.prior_count == 4
        assert mapper.prior_count == 4

        instance = mapper.instance_from_unit_vector([0.1, 0.2, 0.3, 0.4])

        assert len(instance.list_object.ls) == 2
        assert instance.list_object.ls[0].one == 0.1
        assert instance.list_object.ls[0].two == 0.4
        assert instance.list_object.ls[1].one == 0.3
        assert instance.list_object.ls[1].two == 0.8

    def test_mix_instances_and_models(self):
        mapper = af.ModelMapper()
        mapper.list_object = af.PriorModel(
            MockListClass, ls=[MockClassx2, MockClassx2(1, 2)]
        )

        assert mapper.prior_count == 2

        instance = mapper.instance_from_unit_vector([0.1, 0.2])

        assert len(instance.list_object.ls) == 2
        assert instance.list_object.ls[0].one == 0.1
        assert instance.list_object.ls[0].two == 0.4
        assert instance.list_object.ls[1].one == 1
        assert instance.list_object.ls[1].two == 2


class TestCollectionPriorModel:
    def test_keyword_arguments(self):
        prior_model = af.CollectionPriorModel(
            one=MockClassx2, two=MockClassx2(1, 2)
        )

        assert len(prior_model.direct_prior_model_tuples) == 1
        assert len(prior_model) == 2

        instance = prior_model.instance_for_arguments(
            {prior_model.one.one: 0.1, prior_model.one.two: 0.2}
        )

        assert instance.one.one == 0.1
        assert instance.one.two == 0.2

        assert instance.two.one == 1
        assert instance.two.two == 2

    def test_mix_instances_in_grouped_list_prior_model(self):
        prior_model = af.CollectionPriorModel(
            [MockClassx2, MockClassx2(1, 2)]
        )

        assert len(prior_model.direct_prior_model_tuples) == 1
        assert prior_model.prior_count == 2

        mapper = af.ModelMapper()
        mapper.ls = prior_model

        instance = mapper.instance_from_unit_vector([0.1, 0.2])

        assert len(instance.ls) == 2

        assert instance.ls[0].one == 0.1
        assert instance.ls[0].two == 0.4
        assert instance.ls[1].one == 1
        assert instance.ls[1].two == 2

        assert len(prior_model.prior_class_dict) == 2

    def test_list_in_grouped_list_prior_model(self):
        prior_model = af.CollectionPriorModel([[MockClassx2]])

        assert len(prior_model.direct_prior_model_tuples) == 1
        assert prior_model.prior_count == 2

    def test_list_prior_model_with_dictionary(self, simple_model):
        assert isinstance(simple_model.simple, af.PriorModel)

    def test_override_with_instance(self, simple_model):
        simple_instance = MockClassx2(1, 2)

        simple_model.simple = simple_instance

        assert len(simple_model) == 1
        assert simple_model.simple == simple_instance

    def test_names_of_priors(self):
        collection = af.CollectionPriorModel([af.UniformPrior(), af.UniformPrior()])
        assert collection.name_for_prior(collection[0]) == "0"


@pytest.fixture(name="simple_model")
def make_simple_model():
    return af.CollectionPriorModel({"simple": MockClassx2})


class TestCopy:
    def test_simple(self, simple_model):
        assert simple_model.prior_count > 0
        assert copy.deepcopy(simple_model).prior_count == simple_model.prior_count

    def test_embedded(self, simple_model):
        model = af.CollectionPriorModel(simple=simple_model)
        assert copy.deepcopy(model).prior_count == model.prior_count

    def test_circular(self):
        one = af.PriorModel(MockClassx2)

        one.one = af.PriorModel(MockClassx2)
        one.one.one = one

        # noinspection PyUnresolvedReferences
        assert one.prior_count == one.one.prior_count
        assert copy.deepcopy(one).prior_count == one.prior_count



import numpy as np
import pytest
from scipy import stats

from autofit import graphical as graph
from autofit.graphical import Variable, messages
from autofit.messages.normal import NormalMessage


def normal_loglike(x, centre, precision):
    diff = np.asanyarray(x) - centre
    se = np.square(diff)
    return np.sum(-precision * se - np.log(2 * np.pi) + np.log(precision)) / 2


def normal_loglike_jacobian(x, centre, precision):
    diff = np.asanyarray(x) - centre
    se = np.square(diff)
    loglike = np.sum(-precision * se - np.log(2 * np.pi) + np.log(precision)) / 2
    jac_x = -precision * diff
    jac_c = np.sum(precision * diff)
    jac_p = np.sum(1 / precision - se) / 2

    return loglike, (jac_x, jac_c, jac_p)


def normal_loglike_t(x, centre, precision):
    # Make log transform of precision so that optimisation is unbounded
    _precision = np.exp(precision)
    return normal_loglike(x, centre, _precision)


def normal_loglike_t_jacobian(x, centre, precision):
    # Make log transform of precision so that optimisation is unbounded
    _precision = np.exp(precision)
    val, jac = normal_loglike_jacobian(x, centre, _precision)
    jac_x, jac_c, jac_p = jac
    jac_p *= _precision
    return val, (jac_x, jac_c, jac_p)


def hierarchical_loglike(centre, precision, *xs):
    return normal_loglike(xs, centre, precision)


def hierarchical_loglike_jac(centre, precision, *xs):
    loglike, (jac_x, jac_c, jac_p) = normal_loglike_jacobian(xs, centre, precision)
    return loglike, (jac_c, jac_p) + tuple(jac_x)


def hierarchical_loglike_t(centre, logt, *xs):
    return normal_loglike_t(xs, centre, logt)


def hierarchical_loglike_t_jac(centre, precision, *xs):
    loglike, (jac_x, jac_c, jac_p) = normal_loglike_t_jacobian(xs, centre, precision)
    return loglike, (jac_c, jac_p) + tuple(jac_x)


n = 10


@pytest.fixture(name="centres")
def make_centres():
    np.random.seed(1)

    mu = 0.5
    sigma = 0.5
    centre_dist = stats.norm(loc=mu, scale=sigma)

    return centre_dist.rvs(n)


@pytest.fixture(name="widths")
def make_widths():
    np.random.seed(1)

    a = 10
    b = 4000
    precision_dist = stats.invgamma(a, scale=b)

    return precision_dist.rvs(n) ** -0.5


@pytest.fixture(name="model_approx")
def make_model_approx(centres, widths):
    centres_ = [Variable(f"x_{i}") for i in range(n)]
    mu_ = Variable("mu")
    logt_ = Variable("logt")

    centre_likelihoods = [
        NormalMessage(c, w).as_factor(x) for c, w, x in zip(centres, widths, centres_)
    ]
    normal_likelihoods = [
        graph.Factor(
            normal_loglike_t,
            centre,
            mu_,
            logt_,
            factor_jacobian=normal_loglike_t_jacobian,
        )
        for centre in centres_
    ]

    model = graph.utils.prod(centre_likelihoods + normal_likelihoods)

    model_approx = graph.EPMeanField.from_approx_dists(
        model,
        {
            mu_: NormalMessage(0, 10),
            logt_: NormalMessage(0, 10),
            **{x_: NormalMessage(0, 10) for x_ in centres_},
        },
    )

    return model_approx


def test_simple(model_approx, centres):
    laplace = graph.LaplaceOptimiser()
    ep_opt = graph.EPOptimiser(model_approx, default_optimiser=laplace)
    new_approx = ep_opt.run(model_approx, max_steps=20)

    mu_ = new_approx.factor_graph.name_variable_dict["mu"]
    logt_ = new_approx.factor_graph.name_variable_dict["logt"]

    assert new_approx.mean_field[mu_].mean == pytest.approx(np.mean(centres), rel=1.0)
    assert new_approx.mean_field[logt_].mean == pytest.approx(
        np.log(np.std(centres) ** -2), rel=1.0
    )


def test_hierarchical(centres, widths):
    centres_ = [Variable(f"x_{i}") for i in range(n)]
    mu_ = Variable("mu")
    logt_ = Variable("logt")

    centre_likelihoods = [
        messages.NormalMessage(c, w).as_factor(x)
        for c, w, x in zip(centres, widths, centres_)
    ]

    hierarchical_factor = graph.Factor(
        hierarchical_loglike_t,
        mu_,
        logt_,
        *centres_,
        factor_jacobian=hierarchical_loglike_t_jac,
    )

    model = graph.utils.prod(centre_likelihoods) * hierarchical_factor

    model_approx = graph.EPMeanField.from_approx_dists(
        model,
        {
            mu_: messages.NormalMessage(0.0, 10.0),
            logt_: messages.NormalMessage(0.0, 10.0),
            **{x_: messages.NormalMessage(0.0, 10.0) for x_ in centres_},
        },
    )

    laplace = graph.LaplaceOptimiser()
    ep_opt = graph.EPOptimiser(model_approx, default_optimiser=laplace)
    new_approx = ep_opt.run(model_approx, max_steps=10)
    print(new_approx)

    mu_ = new_approx.factor_graph.name_variable_dict["mu"]
    logt_ = new_approx.factor_graph.name_variable_dict["logt"]

    assert new_approx.mean_field[mu_].mean == pytest.approx(np.mean(centres), rel=0.2)
    assert new_approx.mean_field[logt_].mean == pytest.approx(
        np.log(np.std(centres) ** -2), rel=0.2
    )


@pytest.fixture(name="data")
def make_data():
    ## Generative model
    np.random.seed(1)

    mu = 2.0
    sigma = 0.5
    centre_dist = stats.norm(loc=mu, scale=sigma)
    mu_logt = 3.0
    sigma_logt = 0.5
    logt_dist = stats.norm(loc=mu_logt, scale=sigma_logt)

    n = 10
    n_samples_avg = 1000

    centres = centre_dist.rvs(n)
    widths = np.exp(logt_dist.rvs(n)) ** -0.5
    n_samples = np.random.poisson(n_samples_avg, size=n)

    sample_dists = [stats.norm(*p) for p in zip(centres, widths)]
    data = [dist.rvs(n) for dist, n in zip(sample_dists, n_samples)]
    return data


def test_full(data):
    samples = {Variable(f"samples_{i}"): sample for i, sample in enumerate(data)}
    x_i_ = [Variable(f"x_{i}") for i in range(n)]
    logt_i_ = [Variable(f"logt_{i}") for i in range(n)]

    mu_x_ = Variable("mu_x")
    logt_x_ = Variable("logt_x")
    mu_logt_ = Variable("mu_logt")
    logt_logt_ = Variable("logt_logt")
    hierarchical_params = (mu_x_, logt_x_, mu_logt_, logt_logt_)

    # Setting up model
    data_loglikes = [
        graph.Factor(
            normal_loglike_t,
            s_,
            x_,
            logt_,
            factor_jacobian=normal_loglike_t_jacobian,
            name=f"normal_{i}",
        )
        for i, (s_, x_, logt_) in enumerate(zip(samples, x_i_, logt_i_))
    ]
    centre_loglikes = [
        graph.Factor(normal_loglike_t, x_, mu_x_, logt_x_) for x_ in x_i_
    ]
    precision_loglikes = [
        graph.Factor(normal_loglike_t, logt_, mu_logt_, logt_logt_) for logt_ in logt_i_
    ]
    priors = [
        messages.NormalMessage(0, 10).as_factor(v, name=f"prior_{v.name}")
        for v in hierarchical_params
    ]
    model = graph.utils.prod(
        data_loglikes + centre_loglikes + precision_loglikes + priors
    )


def test_full_hierachical(data):
    samples = {Variable(f"samples_{i}"): sample for i, sample in enumerate(data)}
    x_i_ = [Variable(f"x_{i}") for i in range(n)]
    logt_i_ = [Variable(f"logt_{i}") for i in range(n)]

    mu_x_ = Variable("mu_x")
    logt_x_ = Variable("logt_x")
    mu_logt_ = Variable("mu_logt")
    logt_logt_ = Variable("logt_logt")
    hierarchical_params = (mu_x_, logt_x_, mu_logt_, logt_logt_)

    # Setting up model
    data_loglikes = [
        graph.Factor(
            normal_loglike_t,
            s_,
            x_,
            logt_,
            factor_jacobian=normal_loglike_t_jacobian,
            name=f"normal_{i}",
        )
        for i, (s_, x_, logt_) in enumerate(zip(samples, x_i_, logt_i_))
    ]
    centre_loglike = graph.Factor(
        hierarchical_loglike_t,
        mu_x_,
        logt_x_,
        *x_i_,
        name="mean_loglike",
        factor_jacobian=hierarchical_loglike_t_jac,
    )
    logt_loglike = graph.Factor(
        hierarchical_loglike_t,
        mu_logt_,
        logt_logt_,
        *logt_i_,
        name="logt_loglike",
        factor_jacobian=hierarchical_loglike_t_jac,
    )
    priors = [
        messages.NormalMessage(0.0, 10.0).as_factor(v, name=f"prior_{v.name}")
        for v in hierarchical_params
    ]

    model = graph.utils.prod(data_loglikes + priors + [centre_loglike, logt_loglike])

    model_approx = graph.EPMeanField.from_approx_dists(
        model,
        {
            **{v: messages.NormalMessage(0.0, 10.0) for v in model.variables},
            **{v: messages.FixedMessage(sample) for v, sample in samples.items()},
        },
    )

    # Mean field approximation
    model_approx = graph.EPMeanField.from_approx_dists(
        model,
        {
            **{v: messages.NormalMessage(0.0, 10.0) for v in model.variables},
            **{v: messages.FixedMessage(sample) for v, sample in samples.items()},
        },
    )

    laplace = graph.LaplaceOptimiser()
    ep_opt = graph.EPOptimiser(model, default_optimiser=laplace)
    new_approx = ep_opt.run(model_approx, max_steps=20)
    new_approx.mean_field.subset(hierarchical_params)

    m = np.mean([np.mean(sample) for sample in data])
    logt = np.mean([np.log(np.std(sample) ** -2) for sample in data])

    assert new_approx.mean_field[mu_x_].mean == pytest.approx(m, rel=1.0)
    assert new_approx.mean_field[mu_logt_].mean == pytest.approx(logt, rel=1.0)



import pytest

import autofit as af


def _make_grid_paths(
        grid_search,
        mapper
):
    jobs = list(
        grid_search.make_jobs(
            mapper,
            None,
            grid_priors=[
                mapper.component.one_tuple.one_tuple_0,
                mapper.component.one_tuple.one_tuple_1,
            ],
        )
    )
    return [
        job.search_instance.paths
        for job in jobs
    ]


@pytest.fixture(
    name="grid_paths"
)
def make_grid_paths(
        grid_search,
        mapper
):
    return _make_grid_paths(
        grid_search,
        mapper
    )


@pytest.fixture(
    name="database_paths"
)
def make_database_paths(
        grid_search,
        mapper,
        session
):
    grid_search.search.paths = af.DatabasePaths(
        session=session,
        name="grid_search"
    )
    grid_search.search.paths.model = mapper

    return _make_grid_paths(
        grid_search,
        mapper
    )



import math
import numpy as np
from typing import Optional

from autoconf import conf

from autofit.non_linear.analysis import Analysis
from autofit.mapper.model import ModelInstance
from autofit.mapper.model_mapper import ModelMapper
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.result import Result
from autofit.non_linear.samples import PDFSamples, Sample, NestSamples

from autofit import exc


def samples_with_log_likelihood_list(
        log_likelihood_list
):
    return [
        Sample(
            log_likelihood=log_likelihood,
            log_prior=0,
            weight=0
        )
        for log_likelihood
        in log_likelihood_list
    ]


class MockAnalysis(Analysis):
    prior_count = 2

    def __init__(self):
        super().__init__()
        self.fit_instances = list()

    def log_likelihood_function(self, instance):
        self.fit_instances.append(instance)
        return [1]

    def visualize(self, paths, instance, during_analysis):
        pass

    def log(self, instance):
        pass


class MockResult(Result):
    def __init__(
            self,
            samples=None,
            instance=None,
            model=None,
            analysis=None,
            search=None,
    ):

        super().__init__(samples, model, search)

        self._instance = instance or ModelInstance()
        self.model = model or ModelMapper()
        self.samples = samples or MockSamples(max_log_likelihood_instance=self.instance)

        self.gaussian_tuples = None
        self.analysis = analysis
        self.search = search
        self.model = model

    def model_absolute(self, absolute):
        return self.model

    def model_relative(self, relative):
        return self.model

    @property
    def last(self):
        return self


class MockSamples(PDFSamples):
    def __init__(
            self,
            model=None,
            sample_list=None,
            max_log_likelihood_instance=None,
            log_likelihood_list=None,
            gaussian_tuples=None,
            unconverged_sample_size=10,
            **kwargs,
    ):

        self._log_likelihood_list = log_likelihood_list

        self.model = model

        sample_list = sample_list or self.default_sample_list

        super().__init__(
            model=model, sample_list=sample_list, unconverged_sample_size=unconverged_sample_size, **kwargs
        )

        self._max_log_likelihood_instance = max_log_likelihood_instance
        self._gaussian_tuples = gaussian_tuples

    @property
    def default_sample_list(self):

        if self._log_likelihood_list is not None:
            log_likelihood_list = self._log_likelihood_list
        else:
            log_likelihood_list = range(3)

        return [
            Sample(
                log_likelihood=log_likelihood,
                log_prior=0.0,
                weight=0.0
            )
            for log_likelihood
            in log_likelihood_list
        ]

    @property
    def log_likelihood_list(self):

        if self._log_likelihood_list is None:
            return super().log_likelihood_list

        return self._log_likelihood_list

    @property
    def max_log_likelihood_instance(self):

        if self._max_log_likelihood_instance is None:

            try:
                return super().max_log_likelihood_instance
            except (KeyError, AttributeError):
                pass

        return self._max_log_likelihood_instance

    def gaussian_priors_at_sigma(self, sigma=None):

        if self._gaussian_tuples is None:
            return super().gaussian_priors_at_sigma(sigma=sigma)

        return self._gaussian_tuples

    def write_table(self, filename):
        pass


class MockNestSamples(NestSamples):

    def __init__(
            self,
            model,
            sample_list=None,
            total_samples=10,
            log_evidence=0.0,
            number_live_points=5,
            time: Optional[float] = None,
    ):

        self.model = model

        if sample_list is None:

            sample_list = [
                Sample(
                    log_likelihood=log_likelihood,
                    log_prior=0.0,
                    weight=0.0
                )
                for log_likelihood
                in self.log_likelihood_list
            ]

        super().__init__(
            model=model,
            sample_list=sample_list,
            time=time
        )

        self._total_samples = total_samples
        self._log_evidence = log_evidence
        self._number_live_points = number_live_points

    @property
    def total_samples(self):
        return self._total_samples

    @property
    def log_evidence(self):
        return self._log_evidence

    @property
    def number_live_points(self):
        return self._number_live_points


class MockSearch(NonLinearSearch):
    def __init__(
            self,
            name="",
            samples=None,
            result=None,
            unique_tag: Optional[str] = None,
            prior_passer=None,
            fit_fast=True,
            sample_multiplier=1,
            save_for_aggregator=False,
            return_sensitivity_results=False,
            **kwargs
    ):

        super().__init__(name=name, unique_tag=unique_tag, **kwargs)

        self.samples = samples or MockSamples()
        self.result = result or MockResult(samples=samples)

        if prior_passer is not None:
            self.prior_passer = prior_passer

        self.fit_fast = fit_fast
        self.sample_multiplier = sample_multiplier

        self.save_for_aggregator = save_for_aggregator
        self.return_sensitivity_results = return_sensitivity_results

    @property
    def config_type(self):
        return conf.instance["non_linear"]["mock"]

    @property
    def config_dict_search(self):
        return {}

    def _fit_fast(self, model, analysis, log_likelihood_cap=None):
        class Fitness:
            def __init__(self, instance_from_vector, result):
                self.result = result
                self.instance_from_vector = instance_from_vector

            def __call__(self, vector):
                instance = self.instance_from_vector(vector)

                log_likelihood = analysis.log_likelihood_function(instance)

                if self.result.instance is None:
                    self.result.instance = instance

                # Return Chi squared
                return -2 * log_likelihood

        if self.save_for_aggregator:
            analysis.save_attributes_for_aggregator(paths=self.paths)

        fitness_function = Fitness(model.instance_from_vector, result=self.result)
        fitness_function(model.prior_count * [0.8])

        return fitness_function.result

    def _fit(self, model, analysis, log_likelihood_cap=None):

        if self.fit_fast:
            result = self._fit_fast(model=model, analysis=analysis)
            return result

        if model.prior_count == 0:
            raise AssertionError("There are no priors associated with the model!")
        if model.prior_count != len(model.unique_prior_paths):
            raise AssertionError(
                "Prior count doesn't match number of unique prior paths"
            )
        index = 0
        unit_vector = model.prior_count * [0.5]
        while True:
            try:
                instance = model.instance_from_unit_vector(unit_vector)
                fit = analysis.log_likelihood_function(instance)
                break
            except exc.FitException as e:
                unit_vector[index] += 0.1
                if unit_vector[index] >= 1:
                    raise e
                index = (index + 1) % model.prior_count
        samples = MockSamples(
            sample_list=samples_with_log_likelihood_list(self.sample_multiplier * fit),
            model=model,
            gaussian_tuples=[
                (prior.mean, prior.width if math.isfinite(prior.width) else 1.0)
                for prior in sorted(model.priors, key=lambda prior: prior.id)
            ],
        )

        self.paths.save_samples(samples)

        return MockResult(
            model=model,
            samples=samples,
        )

    def perform_update(self, model, analysis, during_analysis):

        if self.samples is not None and not self.return_sensitivity_results:
            self.paths.save_object("samples", self.samples)
            return self.samples

        return MockSamples(
            sample_list=samples_with_log_likelihood_list([1.0, 2.0]),
            gaussian_tuples=[
                (prior.mean, prior.width if math.isfinite(prior.width) else 1.0)
                for prior in sorted(model.priors, key=lambda prior: prior.id)
            ]
        )



from autofit import database as db


def test_is_complete(
        gaussian_1,
        aggregator
):
    assert aggregator.query(
        aggregator.search.is_complete
    ) == [gaussian_1]


def test_is_not_complete(
        gaussian_2,
        aggregator
):
    assert aggregator.query(
        ~aggregator.search.is_complete
    ) == [gaussian_2]


def test_call(
        gaussian_1,
        aggregator
):
    assert aggregator(
        aggregator.search.is_complete
    ) == [gaussian_1]


def test_completed_only(
        gaussian_1,
        session
):
    aggregator = db.Aggregator.from_database(
        'sqlite://',
        completed_only=True
    )
    aggregator.session = session
    assert aggregator == [gaussian_1]


def test_combine(
        aggregator,
        gaussian_1
):
    assert aggregator.query(
        aggregator.search.is_complete & (aggregator.model.centre == 1)
    ) == [gaussian_1]
    assert aggregator.query(
        (~aggregator.search.is_complete) & (aggregator.model.centre == 1)
    ) == []
    assert aggregator.query(
        aggregator.search.is_complete & (aggregator.model.centre == 2)
    ) == []


def test_unique_tag(
        aggregator,
        gaussian_1,
        gaussian_2
):
    assert aggregator.query(
        aggregator.search.unique_tag == "one"
    ) == [gaussian_1]
    assert aggregator.query(
        aggregator.search.unique_tag == "two"
    ) == [gaussian_2]


def test_in(
        aggregator,
        gaussian_1,
        gaussian_2
):
    assert aggregator.query(
        aggregator.search.unique_tag.in_(
            "one something"
        )
    ) == [gaussian_1]
    assert aggregator.query(
        aggregator.search.unique_tag.in_(
            "one something two"
        )
    ) == [gaussian_1, gaussian_2]


def test_contains(
        aggregator,
        gaussian_1,
        gaussian_2
):
    assert aggregator.query(
        aggregator.search.unique_tag.contains(
            "o"
        )
    ) == [
               gaussian_1,
               gaussian_2
           ]
    assert aggregator.query(
        aggregator.search.unique_tag.contains(
            "ne"
        )
    ) == [
               gaussian_1,
           ]



import pytest

import autofit as af
from autofit.mock.mock_model import MockClassx2Tuple
from autofit.mock.mock import MockSearch

@pytest.fixture(name="mapper")
def make_mapper():
    return af.Collection(
        component=af.Model(
            MockClassx2Tuple
        )
    )


@pytest.fixture(name="grid_search")
def make_grid_search(mapper):
    mock_search = MockSearch()
    mock_search.paths = af.DirectoryPaths(name="")
    search = af.SearchGridSearch(
        number_of_steps=10,
        search=mock_search
    )
    return search



import itertools

import pytest

import autofit as af
from autofit import graphical as g
from autofit.graphical import AnalysisFactor
from autofit.graphical.declarative.graph import GraphInfoFormatter
from autofit.mock.mock import MockAnalysis
from autofit.tools.namer import namer


@pytest.fixture(
    autouse=True
)
def reset_namer():
    namer.reset()
    yield
    namer.reset()
    af.ModelObject._ids = itertools.count()


@pytest.fixture(
    name="factor_graph_model"
)
def make_factor_graph_model():
    model_factor_1 = g.AnalysisFactor(
        af.Collection(
            one=af.UniformPrior()
        ),
        MockAnalysis()
    )
    model_factor_2 = g.AnalysisFactor(
        af.Collection(
            one=af.UniformPrior()
        ),
        MockAnalysis()
    )

    return g.FactorGraphModel(
        model_factor_1,
        model_factor_2
    )


@pytest.fixture(
    name="non_trivial_model"
)
def make_non_trivial_model():
    one = af.Model(af.Gaussian)
    two = af.Model(af.Gaussian)

    one.centre = two.centre

    model_factor_1 = g.AnalysisFactor(
        one,
        MockAnalysis()
    )
    model_factor_2 = g.AnalysisFactor(
        two,
        MockAnalysis()
    )

    return g.FactorGraphModel(
        model_factor_1,
        model_factor_2
    )


@pytest.fixture(
    name="factor_graph"
)
def make_factor_graph(
        factor_graph_model
):
    return factor_graph_model.graph


def test_factors_with_type(
        factor_graph
):
    factor_type = AnalysisFactor
    factors = factor_graph._factors_with_type(
        factor_type
    )
    assert len(factors) == 2
    for factor in factors:
        assert isinstance(
            factor,
            AnalysisFactor
        )


@pytest.fixture(
    name="prior_factor"
)
def make_prior_factor(
        factor_graph
):
    return factor_graph.prior_factors[0]


@pytest.fixture(
    name="analysis_factor"
)
def make_analysis_factor(
        factor_graph
):
    return factor_graph.analysis_factors[0]


@pytest.fixture(
    name="declarative_graph_output"
)
def make_declarative_graph_output(
        factor_graph
):
    return GraphInfoFormatter(
        factor_graph
    )



import pytest

import autofit as af
from autofit.mock.mock_model import MockClassx2, MockClassx4

names = ["one", "two", "three"]


@pytest.fixture(name="collection")
def make_collection():
    return af.Collection({name: af.PriorModel(MockClassx2) for name in names})


def test_prior_count(collection):
    assert collection.prior_count == 6


@pytest.mark.parametrize("name", names)
def test_children(collection, name):
    assert getattr(collection, name).prior_count == 2


def test_replace(collection):
    collection.one = af.PriorModel(MockClassx4)

    assert collection.one.prior_count == 4
    assert collection.prior_count == 8



from pathlib import Path

import pytest

import autofit as af
from autoconf.conf import output_path_for_test
from autofit.database.aggregator.scrape import Scraper

from autofit.mock.mock import MockAnalysis, MockSearch

from test_autofit.non_linear.grid.test_optimizer_grid_search import MockOptimizer

output_directory = Path(
    __file__
).parent / "output"


@pytest.fixture(
    name="parent_search"
)
@output_path_for_test(
    output_directory
)
def make_parent_search(model_gaussian_x1):
    search = MockSearch(
        name="parent"
    )
    search.paths.model = model_gaussian_x1
    return search


@pytest.fixture(
    name="database_parent_search"
)
def make_database_parent_search(
        session
):
    return af.DynestyStatic(
        session=session
    )


def _make_grid_search(
        mapper,
        parent_search,
        session=None
):
    search = af.SearchGridSearch(
        search=MockOptimizer(
            session=session
        ),
        number_of_steps=2
    )
    search.fit(
        model=mapper,
        analysis=MockAnalysis(),
        grid_priors=[
            mapper.component.one_tuple.one_tuple_0,
            mapper.component.one_tuple.one_tuple_1,
        ],
        parent=parent_search
    )
    return search


@pytest.fixture(
    name="grid_search"
)
def make_grid_search(
        mapper,
        parent_search
):
    search = _make_grid_search(
        mapper,
        parent_search
    )
    search.paths.save_all()
    return search


@pytest.fixture(
    name="database_grid_search"
)
def make_database_grid_search(
        mapper,
        database_parent_search,
        session
):
    return _make_grid_search(
        mapper,
        database_parent_search,
        session=session
    )


class TestMiscombination:
    def test_directory_for_database(
            self,
            parent_search,
            session,
            mapper
    ):
        with pytest.raises(TypeError):
            _make_grid_search(
                mapper,
                parent_search,
                session
            )


class TestDirectory:
    def test_parent_search(
            self,
            grid_search,
            parent_search
    ):
        grid_paths = grid_search.paths
        parent_paths = parent_search.paths

        assert parent_paths is grid_paths.parent
        with open(
                grid_paths._parent_identifier_path
        ) as f:
            assert f.read() == parent_paths.identifier

    def test_is_grid_search(
            self,
            grid_search
    ):
        assert grid_search.paths.is_grid_search


@output_path_for_test(
    output_directory
)
def test_scrape(
        grid_search,
        parent_search,
        model_gaussian_x1,
        session
):
    grid_search.fit(
        model=model_gaussian_x1,
        analysis=MockAnalysis(),
        parent=parent_search,
        grid_priors=[model_gaussian_x1.centre]
    )
    parent_search.fit(
        model=model_gaussian_x1,
        analysis=MockAnalysis()
    )
    parent_search.paths.save_all()

    Scraper(
        directory=output_directory,
        session=session
    ).scrape()

    aggregator = af.Aggregator(session)
    assert list(aggregator.query(
        aggregator.search.id == grid_search.paths.identifier
    ))[0].parent.id == parent_search.paths.identifier
    assert len(aggregator.values("max_log_likelihood")) > 0
    assert list(aggregator.grid_searches())[0].is_complete


@output_path_for_test(
    output_directory
)
def test_incomplete(
        grid_search,
        session
):
    grid_search.save_metadata()

    Scraper(
        directory=output_directory,
        session=session
    ).scrape()

    session.commit()

    aggregator = af.Aggregator(
        session
    )
    aggregator = aggregator(
        aggregator.search.is_complete
    )
    assert len(aggregator) == 0


class TestDatabase:
    def test_parent_search(
            self,
            database_grid_search,
            database_parent_search
    ):
        parent_paths = database_parent_search.paths

        assert parent_paths is database_grid_search.paths.parent
        assert database_grid_search.paths.fit.parent_id == parent_paths.identifier

    def test_is_grid_search(
            self,
            database_grid_search
    ):
        assert database_grid_search.paths.is_grid_search



from os import path
import pytest

import autofit as af

from autofit.mock.mock_model import MockClassx2
from autofit.mock.mock import MockSamples, MockNestSamples

from autofit.text import text_util

text_path = path.join("{}".format(path.dirname(path.realpath(__file__))), "files", "samples")


@pytest.fixture(name="model")
def make_model():
    return af.ModelMapper(mock_class=MockClassx2)


@pytest.fixture(name="samples")
def make_samples(model):
    parameters = [[1.0, 2.0], [1.2, 2.2]]

    log_likelihood_list = [1.0, 0.0]

    return MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=[0.0, 0.0],
            weight_list=log_likelihood_list,
            model=model
        )
    )


def test__results_to_file(samples):
    file_results = path.join(text_path, "model.results")

    text_util.results_to_file(
        samples=samples, filename=file_results
    )

    results = open(file_results)

    line = results.readline()

    assert (
            line
            == "Maximum Log Likelihood                                                                    1.00000000\n"
    )

    line = results.readline()

    assert (
            line
            == "Maximum Log Posterior                                                                     1.00000000\n"
    )


def test__search_summary_to_file(model):
    file_search_summary = path.join(text_path, "search.summary")

    parameters = [[1.0, 2.0], [1.2, 2.2]]

    log_likelihood_list = [1.0, 0.0]

    samples = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=[0.0, 0.0],
            weight_list=log_likelihood_list,
            model=model
        ),
        time=None,
    )

    text_util.search_summary_to_file(samples=samples, log_likelihood_function_time=1.0, filename=file_search_summary)

    results = open(file_search_summary)
    lines = results.readlines()
    assert lines[0] == "Total Samples = 2\n"
    results.close()

    samples = MockNestSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list + [2.0],
            log_prior_list=[1.0, 1.0],
            weight_list=log_likelihood_list,
            model=model
        ),
        total_samples=10,
        time=2,
        number_live_points=1,
        log_evidence=1.0,
    )

    text_util.search_summary_to_file(samples=samples, log_likelihood_function_time=1.0, filename=file_search_summary)

    results = open(file_search_summary)
    lines = results.readlines()
    assert lines[0] == "Total Samples = 10\n"
    assert lines[1] == "Total Accepted Samples = 2\n"
    assert lines[2] == "Acceptance Ratio = 0.2\n"
    assert lines[3] == "Time To Run = 2\n"
    assert lines[4] == "Log Likelihood Function Evaluation Time (seconds) = 1.0"
    results.close()



import pytest

import autofit as af

from autofit import database as db


@pytest.fixture(
    name="gaussian_1"
)
def make_gaussian_1():
    return db.Fit(
        id="gaussian_1",
        instance=af.Gaussian(
            centre=1
        ),
        info={"info": 1},
        is_complete=True,
        unique_tag="one"
    )


@pytest.fixture(
    name="gaussian_2"
)
def make_gaussian_2():
    return db.Fit(
        id="gaussian_2",
        instance=af.Gaussian(
            centre=2
        ),
        info={"info": 2},
        is_complete=False,
        unique_tag="two"
    )


@pytest.fixture(
    autouse=True
)
def add_to_session(
        gaussian_1,
        gaussian_2,
        session
):
    session.add_all([
        gaussian_1,
        gaussian_2
    ])
    session.commit()



from itertools import chain, count
from typing import Optional, Tuple, Dict, Set
import operator
from abc import ABC, abstractmethod
from functools import wraps, reduce
from math import sqrt

import numpy as np
from autoconf import cached_property

from autofit.mapper.model_object import ModelObject


class Plate:
    _ids = count()

    def __init__(self, name: Optional[str] = None):
        """
        Represents a dimension, such as number of observations, features or dimensions

        Parameters
        ----------
        name
            The name of this dimension
        """
        self.id = next(self._ids)
        self.name = name or f"plate_{self.id}"

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name})"

    def __eq__(self, other):
        return isinstance(other, Plate) and self.id == other.id

    def __hash__(self):
        return self.id

    def __lt__(self, other):
        return self.id < other.id

    def __gt__(self, other):
        return self.id > other.id


class Variable(ModelObject):
    # __slots__ = ("name", "plates")

    def __init__(self, name: str = None, *plates: Plate, id_=None):
        """
        Represents a variable in the problem. This may be fixed data or some coefficient
        that we are optimising for.

        Parameters
        ----------
        name
            The name of this variable
        plates
            Representation of the dimensions of this variable
        """
        self.plates = plates
        super().__init__(id_=id_)
        self.name = name or f"{self.__class__.__name__.lower()}_{self.id}"

    def __repr__(self):
        args = ", ".join(chain([self.name], map(repr, self.plates)))
        return f"{self.__class__.__name__}({args})"

    def __hash__(self):
        return self.id

    def __len__(self):
        return len(self.plates)

    def __str__(self):
        return self.name

    @property
    def ndim(self) -> int:
        """
        How many dimensions does this variable have?
        """
        return len(self.plates)


def variables(*vals):
    """Helper function for making multiple variable objects

    Example
    -------
    x_, a_, b_, y_, z_ = variables("x, a, b", "y, z")
    """
    for val in vals:
        for v in val.split(","):
            yield Variable(v.strip())


# This allows us to treat the class FactorValue as a variable
# that allows us to keep track of the FactorValue vs deterministic
# values when calculating gradients and jacobians
class VariableMetaClass(type, Variable):
    def __new__(cls, clsname, bases, attrs):
        newcls = super().__new__(cls, clsname, bases, attrs)
        Variable.__init__(newcls, clsname)
        return newcls


class FactorValue(np.ndarray, metaclass=VariableMetaClass):
    def __new__(cls, input_array, deterministic_values=None):
        obj = np.asarray(input_array).view(cls)
        obj.deterministic_values = deterministic_values or {}
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.deterministic_values = getattr(obj, "deterministic_values", None)

    @property
    def log_value(self) -> np.ndarray:
        if self.shape:
            return self.base
        else:
            return self.item()

    def __getitem__(self, index) -> np.ndarray:
        if index is type(self):
            return self
        elif isinstance(index, Variable):
            return self.deterministic_values[index]
        else:
            return super().__getitem__(index)

    def keys(self):
        return self.deterministic_values.keys()

    def values(self):
        return self.deterministic_values.values()

    def items(self):
        return self.deterministic_values.items()

    deterministic_variables = property(keys)

    def __repr__(self):
        r = np.ndarray.__repr__(self)
        return r[:-1] + ", " + repr(self.deterministic_values) + ")"

    def to_dict(self):
        return VariableData({FactorValue: self.base, **self.deterministic_values})


def broadcast_plates(
    value: np.ndarray,
    in_plates: Tuple[Plate, ...],
    out_plates: Tuple[Plate, ...],
    reducer: np.ufunc = np.sum,
) -> np.ndarray:
    """
    Extract the indices of a collection of plates then match
    the shape of the data to that shape.

    Parameters
    ----------
    value
        A value to broadcast
    in_plates
        Plates representing the dimensions of the values
    out_plates
        Plates representing the output dimensions
    reducer
        function to reduce excess plates over, default np.sum
        must take axis as keyword argument


    Returns
    -------
    The value reshaped to match the plates
    """
    n_in = len(in_plates)
    n_out = len(out_plates)
    shift = np.ndim(value) - n_in
    if not (0 <= shift <= 1):
        raise ValueError("dimensions of value incompatible with passed plates")

    in_axes = list(range(shift, n_in + shift))
    out_axes = []
    k = n_out + shift

    for plate in in_plates:
        try:
            out_axes.append(out_plates.index(plate) + shift)
        except ValueError:
            out_axes.append(k)
            k += 1

    moved_value = np.moveaxis(
        np.expand_dims(value, tuple(range(n_in + shift, k))),
        in_axes,
        out_axes,
    )
    return reducer(moved_value, axis=tuple(range(n_out + shift, k)))


def _get_variable_data_class(data):
    # So that these methods work with standard python dictionaries
    return type(data) if isinstance(data, VariableData) else VariableData


def _unary_op(op):
    @wraps(op)
    def __op__(self):
        cls = _get_variable_data_class(self)
        return cls({k: op(val) for k, val in self.items()})

    return __op__


def _binary_op(op, ravel=False):
    if ravel:

        @wraps(op)
        def __op__(self, other):
            cls = _get_variable_data_class(self)
            if isinstance(other, FactorValue):
                other = other.to_dict()

            if isinstance(other, dict):
                return cls(
                    {
                        k: op(np.ravel(self[k]), np.ravel(other[k]))
                        for k in self.keys() & other.keys()
                    }
                )
            elif isinstance(other, VariableLinearOperator):
                return op(dict(self), other)
            else:
                return cls({k: op(val, other) for k, val in self.items()})

    else:

        @wraps(op)
        def __op__(self, other):
            cls = _get_variable_data_class(self)
            if isinstance(other, FactorValue):
                other = other.to_dict()

            if isinstance(other, dict):
                return cls(
                    {k: op(self[k], other[k]) for k in self.keys() & other.keys()}
                )
            elif isinstance(other, VariableLinearOperator):
                return op(dict(self), other)
            else:
                return cls({k: op(val, other) for k, val in self.items()})

    return __op__


def rmul(x, y):
    return y * x


def rtruediv(x, y):
    return y / x


class VariableData(Dict[Variable, np.ndarray]):
    var_norm = _unary_op(np.linalg.norm)
    var_det = _unary_op(np.linalg.det)
    var_max = _unary_op(np.max)
    var_min = _unary_op(np.min)
    var_sum = _unary_op(np.sum)
    var_all = _unary_op(np.all)
    var_any = _unary_op(np.any)
    var_prod = _unary_op(np.prod)
    var_isfinite = _unary_op(np.isfinite)

    __abs__ = _unary_op(operator.abs)
    __pos__ = _unary_op(operator.pos)
    __neg__ = _unary_op(operator.neg)
    __lt__ = _binary_op(operator.lt)
    __le__ = _binary_op(operator.le)
    __eq__ = _binary_op(operator.eq)
    __ne__ = _binary_op(operator.ne)
    __gt__ = _binary_op(operator.gt)
    __ge__ = _binary_op(operator.ge)
    __and__ = _binary_op(operator.and_)
    __and__ = _binary_op(operator.or_)

    __add__ = _binary_op(operator.add)
    __radd__ = _binary_op(operator.add)
    __sub__ = _binary_op(operator.sub)
    __mul__ = _binary_op(operator.mul)
    __rmul__ = _binary_op(rmul)
    __truediv__ = _binary_op(operator.truediv)
    __rtruediv__ = _binary_op(rtruediv)
    __pow__ = _binary_op(operator.pow)

    abs = __abs__
    neg = __neg__
    sub = __sub__
    add = __add__
    mul = __mul__
    div = __truediv__
    var_dot = _binary_op(np.dot, ravel=True)

    @property
    def T(self):
        return type(self)({k: val.T for k, val in self.items()})

    @property
    def shapes(self):
        return {k: np.shape(val) for k, val in self.items()}

    @property
    def sizes(self):
        return {k: np.size(val) for k, val in self.items()}

    @property
    def size(self):
        return sum(np.size(val) for val in self.values())

    def ravel(self):
        cls = _get_variable_data_class(self)
        return cls({k: np.ravel(val) for k, val in self.items()})

    def map(self, func, *args, **kwargs):
        cls = _get_variable_data_class(self)
        return cls(
            (k, func(val, *(arg[k] for arg in args), **kwargs))
            for k, val in self.items()
        )

    def reduce(self, func):
        return reduce(func, self.values())

    def mapreduce(self, func, op):
        return VariableData.map(self, func).reduce(op)

    def subset(self, variables):
        cls = _get_variable_data_class(self)
        return cls((v, self[v]) for v in variables)

    def sum(self) -> float:
        return sum(VariableData.var_sum(self).values())

    def prod(self) -> float:
        return VariableData.reduce(VariableData.var_prod(self).values(), operator.mul)

    def all(self) -> bool:
        return all(VariableData.var_all(self).values())

    def any(self) -> bool:
        return any(VariableData.var_all(self).values())

    def det(self) -> float:
        return VariableData.var_det(self).reduce(operator.mul)

    def log_det(self) -> float:
        return VariableData.mapreduce(self, np.linalg.logdet, operator.add)

    def max(self) -> float:
        return max(VariableData.var_max(self).values())

    def min(self) -> float:
        return min(VariableData.var_min(self).values())

    def dot(self, other) -> float:
        return VariableData.var_dot(self, other).sum()

    def norm(self) -> float:
        return sqrt(VariableData.dot(self, self))

    def vecnorm(self, ord: Optional[float] = None) -> float:
        if ord:
            absval = VariableData.abs(self)
            if ord == np.Inf:
                return absval.max()
            elif ord == -np.Inf:
                return absval.min()
            else:
                return (absval ** ord).sum() ** (1.0 / ord)
        else:
            return VariableData.norm(self)

    def __repr__(self):
        name = type(self).__name__
        data_repr = dict.__repr__(self)
        return f"{name}({data_repr})"

    def merge(self, other):
        return VariableData({**self, **other})


class VariableLinearOperator(ABC):
    """Implements the functionality of a linear operator acting
    on a dictionary of values indexed by `Variable` objects
    """

    @abstractmethod
    def __mul__(self, x: VariableData) -> VariableData:
        pass

    @abstractmethod
    def __rtruediv__(self, x: VariableData) -> VariableData:
        pass

    @abstractmethod
    def __rmul__(self, x: VariableData) -> VariableData:
        pass

    @abstractmethod
    def ldiv(self, x: VariableData) -> VariableData:
        pass

    @property
    @abstractmethod
    def variables(self) -> VariableData:
        pass

    def dot(self, x):
        return self * x

    __matmul__ = dot

    def inv(self) -> "InverseVariableOperator":
        return InverseVariableOperator(self)

    def quad(self, M: VariableData) -> VariableData:
        return (M * self).T * self

    def invquad(self, M: VariableData) -> VariableData:
        return (M / self).T / self

    @abstractmethod
    def update(self, *args: Tuple[VariableData, VariableData]):
        pass

    def lowrankupdate(self, *values: VariableData):
        return self.update(*((value, value) for value in values()))

    def lowrankdowndate(self, *values: VariableData):
        return self.update(*((value, VariableData.neg(value)) for value in values()))

    def blocks(self):
        return self.to_block().blocks()


class InverseVariableOperator(VariableLinearOperator):
    def __init__(self, op):
        self.operator = op

    def __mul__(self, x: VariableData) -> VariableData:
        return self.operator.ldiv(x)

    def __rtruediv__(self, x: VariableData) -> VariableData:
        return x * self.operator

    def __rmul__(self, x: VariableData) -> VariableData:
        return x / self.operator

    def ldiv(self, x: VariableData) -> VariableData:
        return self * x

    def quad(self, M: VariableData) -> VariableData:
        return self.operator.invquad(M)

    def invquad(self, M: VariableData) -> VariableData:
        return self.operator.quad(M)

    def inv(self) -> VariableLinearOperator:
        return self.operator

    @property
    def variables(self) -> Set[Variable]:
        return self.operator.variables

    @property
    def is_diagonal(self):
        return self.operator.is_diagonal

    @cached_property
    def log_det(self):
        return -self.operator.log_det

    def update(self, *args: Tuple[VariableData, VariableData]):
        # apply Sherman-Morrison formulat
        A = self.operator
        for (u, v) in args:
            A1u = A * u
            A1v = v * A
            vTA1u = -A1u.dot(v)
            A = A.update(A1u, A1v * vTA1u)

        return type(self)(A)

    def diagonalupdate(self, d: VariableData):
        A = self.operator.diagonalupdate(d ** -1)
        return type(self)(A)

    def to_full(self) -> "VariableFullOperator":
        full_op = self.operator.to_full()
        M = np.linalg.inv(full_op.operator.to_dense())
        return full_op.from_dense(M, full_op.param_shapes)

    def diagonal(self) -> VariableData:
        full_op = self.to_full()
        diag = full_op.operator.to_dense().diagonal()
        return full_op.param_shapes.unflatten(diag)

    def to_block(self) -> "VariableOperator":
        return self.to_full().to_block()



import pytest

from autofit.database import query as q


@pytest.fixture(
    name="less_than"
)
def make_less_than():
    return q.V(
        "<", 1
    )


@pytest.fixture(
    name="greater_than"
)
def make_greater_than():
    return q.V(
        ">", 0
    )


@pytest.fixture(
    name="simple_and"
)
def make_simple_and(
        less_than,
        greater_than
):
    return q.Q(
        "a",
        q.And(
            less_than,
            greater_than
        )
    )


@pytest.fixture(
    name="simple_or"
)
def make_simple_or(
        less_than,
        greater_than
):
    return q.Q(
        "a",
        q.Or(
            less_than,
            greater_than
        )
    )


@pytest.fixture(
    name="second_level"
)
def make_second_level(
        less_than,
        greater_than
):
    return q.Q(
        "a",
        q.And(
            less_than,
            q.Q('b', greater_than)
        )
    )



import numpy as np
import pytest

import autofit as af
from autofit import conf
from autofit.mapper.model_object import Identifier
from autofit.mock.mock import MockSamples, MockSearch, MockAnalysis


def set_version(version):
    conf.instance[
        "general"
    ][
        "output"
    ][
        "identifier_version"
    ] = version


@pytest.fixture(
    autouse=True
)
def set_high_version():
    set_version(99)


def test_identifier_version():
    set_version(1)
    identifier = Identifier(af.Gaussian())

    set_version(2)
    assert identifier != Identifier(af.Gaussian())

    assert identifier == Identifier(
        af.Gaussian(),
        version=1
    )


def test_unique_tag_is_used():
    identifier = af.DynestyStatic(
        "name",
        unique_tag="tag"
    ).paths._identifier

    assert "tag" in identifier.hash_list


def test_class_path():
    identifier = Identifier(
        Class,
        version=3
    )
    string, = identifier.hash_list
    assert "test_autofit.database.identifier.test_identifiers.Class" in string


class Class:
    __doc__ = "hello"

    def __init__(self, one=1, two=2, three=3):
        self.one = one
        self.two = two
        self.three = three
        self.four = None

    __identifier_fields__ = ("one", "two")

    def __eq__(self, other):
        return self.one == other.one


class ExcludeClass:
    def __init__(self, one=1, two=2, three=3):
        self.one = one
        self.two = two
        self.three = three

    __exclude_identifier_fields__ = ("three",)


class AttributeClass:
    def __init__(self):
        self.attribute = None


def test_exclude_identifier_fields():
    other = ExcludeClass(
        three=4
    )
    assert Identifier(
        other
    ) == Identifier(
        ExcludeClass()
    )

    other.__exclude_identifier_fields__ = tuple()

    assert Identifier(
        other
    ) != Identifier(
        ExcludeClass()
    )


def test_numpy_array():
    identifier = Identifier(np.array([0]))
    assert identifier.hash_list == []


def test_hash_list():
    identifier = Identifier(Class())
    assert identifier.hash_list == [
        "Class", "one", "1", "two", "2"
    ]


def test_constructor_only():
    attribute = AttributeClass()
    attribute.attribute = 1

    assert Identifier(
        AttributeClass()
    ) == Identifier(
        attribute
    )


def test_exclude_does_no_effect_constructor():
    attribute = AttributeClass()
    attribute.__exclude_identifier_fields__ = tuple()
    attribute.attribute = 1

    assert Identifier(
        AttributeClass()
    ) == Identifier(
        attribute
    )


class PrivateClass:
    def __init__(self, argument):
        self._argument = argument


def test_private_not_included():
    instance = PrivateClass(
        argument="one"
    )
    identifier = str(Identifier(instance))

    instance._argument = "two"
    assert Identifier(instance) == identifier


def test_missing_field():
    instance = Class()
    instance.__identifier_fields__ = ("five",)

    with pytest.raises(
            AssertionError
    ):
        Identifier(
            instance
        )


def test_change_class():
    gaussian_0 = af.Model(
        af.Gaussian,
        normalization=af.UniformPrior(
            lower_limit=1e-6,
            upper_limit=1e6
        )
    )
    gaussian_1 = af.Model(
        af.Gaussian,
        normalization=af.LogUniformPrior(
            lower_limit=1e-6,
            upper_limit=1e6
        )
    )

    assert Identifier(gaussian_0) != Identifier(gaussian_1)


def test_tiny_change():
    # noinspection PyTypeChecker
    instance = Class(
        one=1.0
    )
    identifier = str(Identifier(instance))

    instance.one += 1e-9
    print(instance.one)

    assert identifier == Identifier(instance)


def test_infinity():
    # noinspection PyTypeChecker
    instance = Class(
        one=float("inf")
    )
    str(Identifier(instance))


def test_identifier_fields():
    other = Class(three=4)
    assert Identifier(
        Class()
    ) == Identifier(
        other
    )

    other.__identifier_fields__ = ("one", "two", "three")
    assert Identifier(
        Class()
    ) != Identifier(
        other
    )


def test_unique_tag():
    search = MockSearch()

    search.fit(
        model=af.Collection(),
        analysis=MockAnalysis()
    )

    identifier = search.paths.identifier

    search = MockSearch(unique_tag="dataset")

    search.fit(
        model=af.Collection(),
        analysis=MockAnalysis(),
    )

    assert search.paths.identifier != identifier


def test_prior():
    identifier = af.UniformPrior().identifier
    assert identifier == af.UniformPrior().identifier
    assert identifier != af.UniformPrior(
        lower_limit=0.5
    ).identifier
    assert identifier != af.UniformPrior(
        upper_limit=0.5
    ).identifier


def test_model():
    identifier = af.PriorModel(
        af.Gaussian,
        centre=af.UniformPrior()
    ).identifier
    assert identifier == af.PriorModel(
        af.Gaussian,
        centre=af.UniformPrior()
    ).identifier
    assert identifier != af.PriorModel(
        af.Gaussian,
        centre=af.UniformPrior(
            upper_limit=0.5
        )
    ).identifier


def test_collection():
    identifier = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.UniformPrior()
        )
    ).identifier
    assert identifier == af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.UniformPrior()
        )
    ).identifier
    assert identifier != af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.UniformPrior(
                upper_limit=0.5
            )
        )
    ).identifier


def test_instance():
    identifier = af.CollectionPriorModel(
        gaussian=af.Gaussian()
    ).identifier
    assert identifier == af.CollectionPriorModel(
        gaussian=af.Gaussian()
    ).identifier
    assert identifier != af.CollectionPriorModel(
        gaussian=af.Gaussian(
            centre=0.5
        )
    ).identifier


def test__identifier_description():

    model = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.UniformPrior(lower_limit=0.0, upper_limit=1.0),
            normalization = af.LogUniformPrior(lower_limit=0.001, upper_limit=0.01),
            sigma=af.GaussianPrior(mean=0.5, sigma=2.0, lower_limit=-1.0, upper_limit=1.0),
        )
    )

    identifier = Identifier([model])

    description = identifier.description.splitlines()

    i=0

    assert description[i] == "CollectionPriorModel"; i+=1
    assert description[i] == "item_number"; i+=1
    assert description[i] == "0"; i+=1
    assert description[i] == "gaussian"; i+=1
    assert description[i] == "PriorModel"; i+=1
    assert description[i] == "cls"; i+=1
    assert description[i] == "autofit.example.model.Gaussian"; i+=1
    assert description[i] == "centre"; i+=1
    assert description[i] == "UniformPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "0.0"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "normalization"; i+=1
    assert description[i] == "LogUniformPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "0.001"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "0.01"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "GaussianPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "-1.0"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "mean"; i+=1
    assert description[i] == "0.5"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "2.0"; i+=1


def test__identifier_description__after_model_and_instance():

    model = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.UniformPrior(lower_limit=0.0, upper_limit=1.0),
            normalization = af.LogUniformPrior(lower_limit=0.001, upper_limit=0.01),
            sigma=af.GaussianPrior(mean=0.5, sigma=2.0, lower_limit=-1.0, upper_limit=1.0),
        )
    )

    max_log_likelihood_instance = model.instance_from_prior_medians()

    samples = MockSamples(
        max_log_likelihood_instance=max_log_likelihood_instance,
        gaussian_tuples=[(1.0, 2.0), (3.0, 4.0), (5.0, 6.0)]
    )
    search = MockSearch(prior_passer=af.PriorPasser(sigma=1.0, use_errors=True, use_widths=False))

    result = af.Result(samples=samples, model=model, search=search)

    model.gaussian.centre = result.model.gaussian.centre
    model.gaussian.normalization = result.instance.gaussian.normalization

    identifier = Identifier([model])

    description = identifier.description.splitlines()

    # THIS TEST FAILS DUE TO THE BUG DESCRIBED IN A GITHUB ISSUE.

    i=0

    assert description[i] == "CollectionPriorModel"; i+=1
    assert description[i] == "item_number"; i+=1
    assert description[i] == "0"; i+=1
    assert description[i] == "gaussian"; i+=1
    assert description[i] == "PriorModel"; i+=1
    assert description[i] == "cls"; i+=1
    assert description[i] == "autofit.example.model.Gaussian"; i+=1
    assert description[i] == "centre"; i+=1
    assert description[i] == "GaussianPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "0.0"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "mean"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "2.0"; i+=1
    assert description[i] == "normalization"; i+=1
    assert description[i] == "0.00316228"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "GaussianPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "-1.0"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "mean"; i+=1
    assert description[i] == "0.5"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "2.0"; i+=1


def test__identifier_description__after_take_attributes():

    model = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.UniformPrior(lower_limit=0.0, upper_limit=1.0),
            normalization = af.LogUniformPrior(lower_limit=0.001, upper_limit=0.01),
            sigma=af.GaussianPrior(mean=0.5, sigma=2.0, lower_limit=-1.0, upper_limit=1.0),
        )
    )

    model.take_attributes(source=model)

    identifier = Identifier([model])

    description = identifier.description.splitlines()

    # THIS TEST FAILS DUE TO THE BUG DESCRIBED IN A GITHUB ISSUE.

    i=0

    assert description[i] == "CollectionPriorModel"; i+=1
    assert description[i] == "item_number"; i+=1
    assert description[i] == "0"; i+=1
    assert description[i] == "gaussian"; i+=1
    assert description[i] == "PriorModel"; i+=1
    assert description[i] == "cls"; i+=1
    assert description[i] == "autofit.example.model.Gaussian"; i+=1
    assert description[i] == "centre"; i+=1
    assert description[i] == "UniformPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "0.0"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "normalization"; i+=1
    assert description[i] == "LogUniformPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "0.001"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "0.01"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "GaussianPrior"; i+=1
    assert description[i] == "lower_limit"; i+=1
    assert description[i] == "-1.0"; i+=1
    assert description[i] == "upper_limit"; i+=1
    assert description[i] == "1.0"; i+=1
    assert description[i] == "mean"; i+=1
    assert description[i] == "0.5"; i+=1
    assert description[i] == "sigma"; i+=1
    assert description[i] == "2.0"; i+=1


import random
from abc import ABC, abstractmethod
from typing import Union, Tuple

from autoconf import conf
from autofit import exc
from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.mapper.prior.deferred import DeferredArgument
from autofit.mapper.variable import Variable

epsilon = 1e-14


def assert_within_limits(
        func,
):
    # noinspection PyShadowingNames
    def wrapper(
            self,
            *args,
            ignore_prior_limits=False,
            **kwargs,
    ):
        value = func(
            self, *args, **kwargs
        )
        if not ignore_prior_limits:
            self.assert_within_limits(value)
        return value

    return wrapper


class Prior(Variable, ABC, ArithmeticMixin):
    __database_args__ = (
        "lower_limit",
        "upper_limit",
        "id_"
    )

    def __init__(
            self,
            lower_limit=0.0,
            upper_limit=1.0,
            id_=None
    ):
        """
        An object used to mappers a unit value to an attribute value for a specific
        class attribute.

        Parameters
        ----------
        lower_limit: Float
            The lowest value this prior can return
        upper_limit: Float
            The highest value this prior can return
        """
        super().__init__(
            id_=id_
        )
        self.lower_limit = float(lower_limit)
        self.upper_limit = float(upper_limit)
        if self.lower_limit >= self.upper_limit:
            raise exc.PriorException(
                "The upper limit of a prior must be greater than its lower limit"
            )

    @property
    @abstractmethod
    def factor(self):
        """
        A callable PDF used as a factor in factor graphs
        """

    def assert_within_limits(self, value):
        if conf.instance["general"]["model"]["ignore_prior_limits"]:
            return
        if not (
                self.lower_limit - epsilon <= value <= self.upper_limit + epsilon
        ):
            raise exc.PriorLimitException(
                "The physical value {} for a prior "
                "was not within its limits {}, {}".format(
                    value, self.lower_limit, self.upper_limit
                )
            )

    @staticmethod
    def for_class_and_attribute_name(cls, attribute_name):
        prior_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name]
        )
        return Prior.from_dict(prior_dict)

    @property
    def width(self):
        return self.upper_limit - self.lower_limit

    def random(self) -> float:
        """
        A random value sampled from this prior
        """
        return self.value_for(
            random.random()
        )

    @abstractmethod
    def value_for(self, unit: float) -> float:
        """
        Return a physical value for a value between 0 and 1 with the transformation
        described by this prior.

        Parameters
        ----------
        unit
            A hypercube value between 0 and 1.

        Returns
        -------
        A physical value.
        """

    def instance_for_arguments(self, arguments):
        return arguments[self]

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<{} id={} lower_limit={} upper_limit={}>".format(
            self.__class__.__name__, self.id, self.lower_limit, self.upper_limit
        )

    def __float__(self):
        return self.value_for(0.5)

    @classmethod
    def from_dict(cls, prior_dict: dict) -> Union["Prior", DeferredArgument]:
        """
        Returns a prior from a JSON representation.

        Parameters
        ----------
        prior_dict : dict
            A dictionary representation of a prior including a type (e.g. Uniform) and all constructor arguments.

        Returns
        -------
        An instance of a child of this class.
        """
        if prior_dict["type"] == "Constant":
            return prior_dict["value"]
        if prior_dict["type"] == "Deferred":
            return DeferredArgument()

        from .prior import UniformPrior
        from .prior import LogUniformPrior
        from .prior import GaussianPrior

        prior_type_dict = {
            "Uniform": UniformPrior,
            "LogUniform": LogUniformPrior,
            "Gaussian": GaussianPrior,
        }

        # noinspection PyProtectedMember
        return prior_type_dict[prior_dict["type"]](
            **{
                key: value
                for key, value in prior_dict.items()
                if key not in ("type", "width_modifier", "gaussian_limits")
            }
        )

    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = {
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit,
            "type": self.name_of_class(),
        }
        return prior_dict

    @classmethod
    def name_of_class(cls) -> str:
        """
        A string name for the class, with the prior suffix removed.
        """
        return cls.__name__.replace("Prior", "")

    @property
    def limits(self) -> Tuple[float, float]:
        return self.lower_limit, self.upper_limit



import os
from pathlib import Path
import pytest
import json

import autofit as af
from autofit.mock.mock_model import MockWithTuple

@pytest.fixture(
    name="model_dict"
)
def make_model_dict():
    return {
        "type": "model",
        "class_path": "autofit.example.model.Gaussian",
        "centre": {'lower_limit': 0.0, 'type': 'Uniform', 'upper_limit': 2.0},
        "normalization": {'lower_limit': 0.0, 'type': 'Uniform', 'upper_limit': 1.0},
        "sigma": {'lower_limit': 0.0, 'type': 'Uniform', 'upper_limit': 1.0},
    }


@pytest.fixture(
    name="instance_dict"
)
def make_instance_dict():
    return {
        "type": "instance",
        "class_path": "autofit.example.model.Gaussian",
        "centre": 0.0,
        "normalization": 0.1,
        "sigma": 0.01
    }


@pytest.fixture(
    name="collection_dict"
)
def make_collection_dict(
        model_dict
):
    return {
        "gaussian": model_dict,
        "type": "collection"
    }


@pytest.fixture(
    name="model"
)
def make_model():
    return af.Model(
        af.Gaussian,
        centre=af.UniformPrior(
            upper_limit=2.0
        )
    )


class TestTuple:
    def test_tuple_prior(self):
        tuple_prior = af.TuplePrior()
        tuple_prior.tup_0 = 0
        tuple_prior.tup_1 = 1

        result = af.Model.from_dict(
            tuple_prior.dict()
        )
        assert isinstance(
            result,
            af.TuplePrior
        )

    def test_model_with_tuple(self):
        tuple_model = af.Model(MockWithTuple)
        tuple_model.instance_from_prior_medians()
        model_dict = tuple_model.dict()

        model = af.Model.from_dict(
            model_dict
        )
        instance = model.instance_from_prior_medians()
        assert instance.tup == (0.5, 0.5)


class TestFromDict:
    def test_model_from_dict(
            self,
            model_dict
    ):
        model = af.Model.from_dict(
            model_dict
        )
        assert model.cls == af.Gaussian
        assert model.prior_count == 3
        assert model.centre.upper_limit == 2.0

    def test_instance_from_dict(
            self,
            instance_dict
    ):
        instance = af.Model.from_dict(
            instance_dict
        )
        assert isinstance(
            instance,
            af.Gaussian
        )
        assert instance.centre == 0.0
        assert instance.normalization == 0.1
        assert instance.sigma == 0.01

    def test_collection_from_dict(
            self,
            collection_dict
    ):
        collection = af.Model.from_dict(
            collection_dict
        )
        assert isinstance(
            collection,
            af.Collection
        )
        assert len(collection) == 1


class TestToDict:
    def test_model_priors(
            self,
            model,
            model_dict
    ):
        assert model.dict() == model_dict

    def test_model_floats(
            self,
            instance_dict
    ):
        model = af.Model(
            af.Gaussian,
            centre=0.0,
            normalization=0.1,
            sigma=0.01
        )

        assert model.dict() == instance_dict

    def test_collection(
            self,
            model,
            collection_dict
    ):
        collection = af.Collection(
            gaussian=model
        )
        assert collection.dict() == collection_dict

    def test_collection_instance(
            self,
            instance_dict
    ):
        collection = af.Collection(
            gaussian=af.Gaussian()
        )
        assert collection.dict() == {
            "gaussian": instance_dict,
            "type": "collection"
        }


class TestFromJson:

    def test__from_json(self, model_dict):

        model = af.Model.from_dict(
            model_dict
        )

        model_file = Path(__file__).parent / "model.json"

        try:
            os.remove(model_file)
        except OSError:
            pass

        with open(model_file, "w+") as f:
            json.dump(model.dict(), f, indent=4)

        model = af.Model.from_json(file=model_file)

        assert model.cls == af.Gaussian
        assert model.prior_count == 3
        assert model.centre.upper_limit == 2.0

        os.remove(model_file)


import pytest

import autofit as af
from autofit import database as db


def _make_children(
        grid_id
):
    return [
        db.Fit(
            id=f"child_{grid_id}_{i}",
            instance=af.Gaussian(
                centre=i
            ),
            model=af.Model(
                af.Gaussian,
                centre=float(-i)
            ),
            max_log_likelihood=grid_id + i
        )
        for i in range(10)
    ]


@pytest.fixture(
    name="children"
)
def make_children():
    return _make_children(1)


@pytest.fixture(
    name="grid_fit"
)
def make_grid_fit(children):
    return db.Fit(
        id="grid",
        unique_tag="grid_fit_1",
        is_grid_search=True,
        children=children,
        instance=af.Gaussian(
            centre=1
        )
    )


@pytest.fixture(
    name="grid_fit_2"
)
def make_grid_fit_2():
    return db.Fit(
        id="grid2",
        unique_tag="grid_fit_2",
        is_grid_search=True,
        children=_make_children(2),
        instance=af.Gaussian(
            centre=2
        )
    )


@pytest.fixture(
    autouse=True
)
def add_to_session(
        grid_fit,
        session
):
    session.add(
        grid_fit
    )
    session.flush()


def test_convert_prior():
    assert float(
        af.UniformPrior(
            lower_limit=0,
            upper_limit=1
        )
    ) == 0.5


def test_cell_aggregator(
        aggregator
):
    assert aggregator.grid_searches().cell_number(
        2
    ).values(
        "id"
    ) == [
               "child_1_3"
           ]


def test_model_order_no():
    model_1 = af.Model(
        af.Gaussian,
        centre=1.0
    )
    model_2 = af.Model(
        af.Gaussian,
        centre=2.0
    )

    assert model_1.order_no < model_2.order_no


def test_negative():
    model_1 = af.Model(
        af.Gaussian,
        centre=-3.0
    )
    model_2 = af.Model(
        af.Gaussian,
        centre=2.0
    )
    assert model_1.order_no < model_2.order_no


def test_model_order_no_complicated():
    model_1 = af.Model(
        af.Gaussian,
        centre=1.0,
        normalization=af.UniformPrior(0.0, 1.0)
    )
    model_2 = af.Model(
        af.Gaussian,
        centre=2.0,
        normalization=af.UniformPrior(0.0, 0.5)
    )
    model_3 = af.Model(
        af.Gaussian,
        centre=2.0,
        normalization=af.UniformPrior(0.0, 1.0)
    )

    assert model_1.order_no < model_2.order_no < model_3.order_no


def test_grid_search_best_fits(
        aggregator
):
    best_fit = aggregator.grid_searches().best_fits()
    assert isinstance(
        best_fit,
        db.GridSearchAggregator
    )
    assert best_fit[0].max_log_likelihood == 10


def test_multiple_best_fits(
        aggregator,
        session,
        grid_fit_2
):
    session.add(grid_fit_2)
    session.commit()
    best_fit = aggregator.grid_searches().best_fits()
    assert best_fit.values(
        "max_log_likelihood"
    ) == [10, 11]

    for parent, fit in zip(
            aggregator.grid_searches(),
            best_fit
    ):
        assert parent is fit.parent


def test_grid_search(
        aggregator,
        grid_fit
):
    result, = aggregator.query(
        aggregator.search.is_grid_search
    ).fits

    assert result is grid_fit


def test_grid_searches(
        aggregator
):
    aggregator = aggregator.grid_searches()
    assert isinstance(
        aggregator,
        db.GridSearchAggregator
    )


class TestChildren:
    def test_simple(
            self,
            aggregator,
            children
    ):
        assert aggregator.grid_searches().children().fits == children

    def test_query_after(
            self,
            aggregator
    ):
        results = aggregator.grid_searches().children().query(
            af.Query().centre <= 5
        ).fits
        assert len(results) == 6

    def test_query_grid_unique_tag(
            self,
            aggregator,
            grid_fit,
            grid_fit_2,
            session
    ):
        session.add(
            grid_fit_2
        )
        session.flush()

        grid_agg = aggregator.grid_searches()
        assert grid_agg[0].unique_tag == "grid_fit_1"

        one = grid_agg.query(
            grid_agg.search.unique_tag == "grid_fit_1"
        )
        assert len(one) == 1
        assert one[0].unique_tag == "grid_fit_1"

    def test_query_grid_unique_tag_none(
            self,
            aggregator,
            grid_fit,
            grid_fit_2,
            session
    ):
        grid_fit_2.unique_tag = None
        session.add(
            grid_fit_2
        )
        session.flush()

        grid_agg = aggregator.grid_searches()
        none = grid_agg.query(
            grid_agg.search.unique_tag == "grid_fit_3"
        )
        assert len(none) == 0

    def test_query_before(
            self,
            aggregator,
            grid_fit,
            grid_fit_2,
            session
    ):
        session.add(
            grid_fit_2
        )
        session.flush()

        parent_aggregator = aggregator.query(
            aggregator.search.is_grid_search & (aggregator.model.centre == 1)
        )

        result, = parent_aggregator.fits

        assert result is grid_fit

        child_aggregator = parent_aggregator.grid_searches().children()

        results = child_aggregator.fits
        assert len(results) == 10

        results = aggregator.query(
            aggregator.search.is_grid_search & (aggregator.model.centre == 2)
        ).grid_searches().children().fits
        assert len(results) == 10



from os import path

import pytest

import autofit as af


@pytest.fixture(name="aggregator")
def make_path_aggregator(session):
    fits = [
        af.db.Fit(
            id="complete",
            is_complete=True
        ),
        af.db.Fit(
            id="incomplete",
            is_complete=False
        )
    ]
    for i, fit in enumerate(fits):
        fit["dataset"] = {
            "name": "dataset"
        }
        fit["pipeline"] = f"pipeline{i}"
    session.add_all(fits)
    session.flush()
    return af.Aggregator(session)


@pytest.fixture(name="aggregator_directory")
def make_aggregator_directory():
    directory = path.dirname(path.realpath(__file__))

    return path.join(f"{directory}", "..", "tools", "files", "aggregator")



from pathlib import Path


def test_sensitivity(
        sensitivity
):
    results = sensitivity.run()
    assert len(results) == 8

    path = Path(
        sensitivity.search.paths.output_path
    ) / "results.csv"
    assert path.exists()
    with open(path) as f:
        assert next(
            f) == 'index,centre,normalization,sigma,log_likelihood_base,log_likelihood_perturbed,log_likelihood_difference\n'
        assert next(f) == '0,0.25,0.25,0.25,2.0,2.0,0.0\n'
        assert next(f) == '1,0.25,0.25,0.75,2.0,2.0,0.0\n'



from random import seed

import matplotlib.pyplot as plt
import numpy as np
import pytest

import autofit as af
from autofit import graphical as g
from test_autofit.graphical.gaussian.model import Analysis

x = np.arange(200)
n = 1

should_plot = False


@pytest.fixture(name="centre_model")
def make_centre_model():
    return g.HierarchicalFactor(
        af.GaussianPrior,
        mean=af.GaussianPrior(mean=100, sigma=10),
        sigma=af.GaussianPrior(mean=10, sigma=5),
    )


def test_embedded_priors(centre_model):
    assert isinstance(centre_model.random_instance().value_for(0.5), float)


def test_hierarchical_factor(centre_model):
    centre_model.add_drawn_variable(af.GaussianPrior(100, 10))
    factor = centre_model.factors[0]

    assert len(factor.priors) == 3

    laplace = g.LaplaceOptimiser()

    gaussian = factor.optimise(laplace, max_steps=10)
    assert gaussian.instance_from_prior_medians().drawn_prior.mean() == pytest.approx(
        100, abs=1
    )


@pytest.fixture(name="centres")
def make_centre(centre_model):
    seed(1)
    centres = list()
    for _ in range(n):
        centres.append(centre_model.random_instance().value_for(0.5))
    return centres


@pytest.fixture(name="data")
def generate_data(centres):
    data = []
    for centre in centres:
        gaussian = af.Gaussian(
            centre=centre,
            normalization=20,
            sigma=5,
        )

        data.append(gaussian(x))
    return data


def test_generate_data(data):
    if should_plot:
        for gaussian in data:
            plt.plot(x, gaussian)
        plt.show(block=False)


def test_model_factor(data, centres):
    y = data[0]
    centre_argument = af.GaussianPrior(mean=50, sigma=20)
    prior_model = af.PriorModel(
        af.Gaussian, centre=centre_argument, normalization=20, sigma=5
    )
    factor = g.AnalysisFactor(prior_model, analysis=Analysis(x=x, y=y))
    laplace = g.LaplaceOptimiser()

    gaussian = factor.optimise(laplace, max_steps=10)
    assert gaussian.centre.mean == pytest.approx(centres[0], abs=0.1)


def test_full_fit(centre_model, data, centres):
    graph = g.FactorGraphModel()
    for i, y in enumerate(data):
        prior_model = af.PriorModel(
            af.Gaussian,
            centre=af.GaussianPrior(mean=100, sigma=20),
            normalization=20,
            sigma=5,
        )
        graph.add(g.AnalysisFactor(prior_model, analysis=Analysis(x=x, y=y)))
        centre_model.add_drawn_variable(prior_model.centre)

    graph.add(centre_model)

    optimiser = g.LaplaceOptimiser()

    collection = graph.optimise(optimiser, max_steps=10).model

    # TODO I don't know what's going on here?
    # pred_centre = (
    #     collection.HierarchicalFactor0.distribution_model.instance_from_prior_medians().mean
    # )
    # (centre,) = centres
    # pred_centre == pytest.approx(centre, rel=0.1)

    # for gaussian, centre in zip(collection.with_prefix("AnalysisFactor"), centres):
    #     assert gaussian.instance_from_prior_medians().centre == pytest.approx(
    #         centre, abs=0.1
    #     )



import autofit as af
from autofit.non_linear.grid import sensitivity as s


def test_lists(sensitivity):
    assert len(list(sensitivity._perturbation_instances)) == 8


def test_tuple_step_size(sensitivity):
    sensitivity.number_of_steps = (2, 2, 4)

    assert len(sensitivity._lists) == 16


def test_labels(sensitivity):
    labels = list(sensitivity._labels)
    assert labels == [
        "centre_0.25_normalization_0.25_sigma_0.25",
        "centre_0.25_normalization_0.25_sigma_0.75",
        "centre_0.25_normalization_0.75_sigma_0.25",
        "centre_0.25_normalization_0.75_sigma_0.75",
        "centre_0.75_normalization_0.25_sigma_0.25",
        "centre_0.75_normalization_0.25_sigma_0.75",
        "centre_0.75_normalization_0.75_sigma_0.25",
        "centre_0.75_normalization_0.75_sigma_0.75",
    ]


def test_searches(sensitivity):
    assert len(list(sensitivity._searches)) == 8


def test_perform_job(job):
    result = job.perform()
    assert isinstance(result, s.JobResult)
    assert isinstance(result.perturbed_result, af.Result)
    assert isinstance(result.result, af.Result)


def test_job_paths(
        job,
        search
):
    output_path = search.paths.output_path
    assert job.perturbed_search.paths.output_path == f"{output_path}/[perturbed]"
    assert job.search.paths.output_path == f"{output_path}/[base]"



from collections import defaultdict
from itertools import repeat
from typing import Optional, Dict, Tuple, Any, List, Iterator

import numpy as np
from scipy.optimize import minimize, OptimizeResult, least_squares, approx_fprime

from autofit.graphical.expectation_propagation import (
    EPMeanField,
    AbstractFactorOptimiser,
)
from autofit.graphical.factor_graphs import transform as t
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.mean_field import MeanField, FactorApproximation, Status
from autofit.graphical.utils import propagate_uncertainty, FlattenArrays, OptResult
from autofit.mapper.operator import LinearOperator
from autofit.mapper.variable import Variable, VariableData
from autofit.messages.fixed import FixedMessage

ArraysDict = Dict[Variable, np.ndarray]


class OptFactor:
    """ """

    def __init__(
            self,
            factor: Factor,
            param_shapes: FlattenArrays,
            fixed_kws: Optional[Dict[str, np.ndarray]] = None,
            model_dist: Optional[MeanField] = None,
            transform: Optional[LinearOperator] = None,
            bounds: Optional[Dict[str, Tuple[float, float]]] = None,
            method: str = "L-BFGS-B",
            jac=False,
            tol=None,
            options=None,
            callback=None,
            constraints=None,
    ):
        self.factor = factor
        self.param_shapes = param_shapes
        self._model_dist = model_dist

        self.jac = jac

        self.transform = t.identity_transform if transform is None else transform
        self.param_bounds = bounds
        self.free_vars = tuple(self.param_shapes.keys())
        self.deterministic_variables = self.factor.deterministic_variables

        self.sign = 1
        self.fixed_kws = fixed_kws or {}

        meth = method.lower()
        # method needs to return Hessian information.
        if meth not in ("bfgs", "l-bfgs-b"):
            raise ValueError("Unknown solver %s" % method)

        self.method = meth

        if bounds:
            # TODO check that this is correct for composite
            # distributions e.g. NormalGammaMessage
            bounds = [
                b
                for k, s in self.param_shapes.items()
                for bound in bounds[k]
                for b in repeat(bound, np.prod(s, dtype=int))
            ]
            self.bounds = self.transform.transform_bounds(bounds)
        else:
            self.bounds = bounds

        self.default_kws = {
            "jac": self.jac,
            "bounds": self.bounds,
            "method": self.method,
            # Might want to automatically specify constraint from messages
            "constraints": constraints,
            "tol": tol,
            "callback": callback,
            "options": options,
        }

    @property
    def model_dist(self):
        if self._model_dist is None:
            raise ValueError("Model dist not defined")
        else:
            return self._model_dist

    @classmethod
    def from_approx(
            cls,
            factor_approx: FactorApproximation,
            transform: Optional[LinearOperator] = None,
    ) -> "OptFactor":
        value_shapes = {}
        fixed_kws = {}
        bounds = {}
        for v in factor_approx.variables:
            dist = factor_approx.model_dist[v]
            if isinstance(dist, FixedMessage):
                fixed_kws[v] = dist.mean
            else:
                value_shapes[v] = dist.shape
                bounds[v] = dist._support

        return cls(
            factor_approx,
            FlattenArrays(value_shapes),
            fixed_kws=fixed_kws,
            model_dist=factor_approx.model_dist,
            transform=transform,
            bounds=bounds,
        )

    def flatten(self, values: Dict[Variable, np.ndarray]) -> np.ndarray:
        x0 = self.param_shapes.flatten(values)
        return x0

    def unflatten(self, x0: np.ndarray) -> Dict[Variable, np.ndarray]:
        values = {**self.param_shapes.unflatten(x0), **self.fixed_kws}
        return values

    def __call__(self, x0):
        values = self.unflatten(self.transform.ldiv(x0))
        return self.sign * np.sum(self.factor(values))

    def func_jacobian(self, x0):
        values = self.unflatten(self.transform.ldiv(x0))
        fval, jval = self.factor.func_jacobian(
            values, self.free_vars, _calc_deterministic=True
        )

        grad = self.flatten(jval) / self.transform
        return self.sign * fval.log_value, self.sign * grad

    def jacobian(self, args):
        return self.func_jacobian(args)[1]

    def numerically_verify_jacobian(self, n_tries=10, eps=1e-6, rtol=1e-3, atol=1e-2):
        x0s = (self.flatten(self.get_random_start()) for _ in range(n_tries))
        return all(
            np.allclose(
                self.jacobian(x0), approx_fprime(x0, self, eps), atol=atol, rtol=rtol
            )
            for x0 in x0s
        )

    def get_random_start(self, arrays_dict: Optional[ArraysDict] = None):
        arrays_dict = arrays_dict or {}
        return {
            v: arrays_dict[v] if v in arrays_dict else self.model_dist[v].sample()
            for v in self.free_vars
        }

    def _parse_result(
            self, result: OptimizeResult, status: Status = Status()
    ) -> OptResult:
        success, messages, flag = status
        success = result.success
        try:
            message = result.message.decode()
        except AttributeError:
            message = result.message
        messages += (
            "optimise.find_factor_mode: "
            f"nfev={result.nfev}, nit={result.nit}, "
            f"status={result.status}, message={message}",
        )

        full_hess_inv = result.hess_inv
        if not isinstance(full_hess_inv, np.ndarray):
            # if optimiser is L-BFGS-B then convert
            # implicit hess_inv into dense matrix
            full_hess_inv = full_hess_inv.todense()

        # make inverse transform back
        M = self.transform
        x = M.ldiv(result.x)
        full_hess_inv = M.ldiv(M.ldiv(full_hess_inv).T)

        mode = {**self.param_shapes.unflatten(x), **self.fixed_kws}
        hess_inv = self.param_shapes.unflatten(full_hess_inv)

        return OptResult(
            mode,
            hess_inv,
            self.sign * result.fun,  # minimized negative logpdf of factor approximation
            full_hess_inv,  # full inverse hessian of optimisation
            result,
            Status(success=success, messages=messages, flag=flag),
        )

    def _minimise(self, arrays_dict, **kwargs):
        x0 = self.transform * self.param_shapes.flatten(arrays_dict)
        opt_kws = {**self.default_kws, **kwargs}
        func = self.func_jacobian if opt_kws["jac"] else self
        return minimize(func, x0, **opt_kws)

    def minimise(
            self,
            arrays_dict: Optional[ArraysDict] = None,
            status: Status = Status(),
            **kwargs,
    ):
        self.sign = 1
        p0 = self.get_random_start(arrays_dict or {})
        res = self._minimise(p0, **kwargs)
        return self._parse_result(res, status=status)

    def maximise(
            self,
            arrays_dict: Optional[Dict[Variable, np.ndarray]] = None,
            status: Status = Status(),
            **kwargs,
    ):
        self.sign = -1
        p0 = self.get_random_start(arrays_dict or {})
        res = self._minimise(p0, **kwargs)
        return self._parse_result(res, status=status)

    minimize = minimise
    maximize = maximise


def update_det_cov(res: OptResult, jacobian: VariableData):
    """Calculates the inv hessian of the deterministic variables

    Note that this modifies res.
    """
    covars = res.hess_inv
    for v, grad in jacobian.items():
        for det, jac in grad.items():
            cov = propagate_uncertainty(covars[v], jac)
            covars[det] = covars.get(det, 0.0) + cov

    return res


class LaplaceFactorOptimiser(AbstractFactorOptimiser):
    def __init__(
            self,
            whiten_optimiser=True,
            transforms=None,
            deltas=None,
            initial_values=None,
            opt_kws=None,
            default_opt_kws=None,
            transform_cls=t.InvCholeskyTransform,
    ):
        super().__init__(deltas=deltas)

        self.whiten_optimiser = whiten_optimiser
        self.initial_values = {}
        if initial_values:
            self.initial_values.update(initial_values)

        self.transforms = defaultdict(lambda: t.identity_transform)
        if transforms:
            self.transforms.update(transforms)

        self.deltas = defaultdict(lambda: 1)
        if deltas:
            self.deltas.update(deltas)

        self.default_opt_kws = default_opt_kws or {}
        self.opt_kws = defaultdict(self.default_opt_kws.copy)
        if opt_kws:
            self.opt_kws.update(opt_kws)

        self.transform_cls = transform_cls

    def optimise(
            self,
            factor: Factor,
            model_approx: EPMeanField,
            status: Optional[Status] = Status(),
    ) -> Tuple[EPMeanField, Status]:

        whiten = self.transforms[factor]
        opt_kws = self.opt_kws[factor]
        start = self.initial_values.get(factor)

        factor_approx = model_approx.factor_approximation(factor)
        opt = OptFactor.from_approx(factor_approx, transform=whiten)
        res = opt.maximise(start, status=status, **opt_kws)

        # Calculate covariance of deterministic values
        # TODO: estimate this Jacobian using Broyden's method
        # https://en.wikipedia.org/wiki/Broyden%27s_method
        value = factor_approx.factor(res.mode)
        res.mode.update(value.deterministic_values)
        jacobian = factor_approx.factor.jacobian(res.mode, opt.free_vars)
        update_det_cov(res, jacobian)

        self.transforms[factor] = self.transform_cls.from_dense(res.full_hess_inv)

        # Project Laplace's approximation
        new_model_dist = factor_approx.model_dist.from_mode_covariance(
            res.mode, res.hess_inv, res.log_norm
        )
        return self.update_model_approx(
            new_model_dist, factor_approx, model_approx, status
        )


LaplaceFactorOptimizer = LaplaceFactorOptimiser


#################################################


def maximise_factor_approx(factor_approx: FactorApproximation, **kwargs):
    """ """
    p0 = {
        v: kwargs.pop(v, factor_approx.model_dist[v].sample(1)[0])
        for v in factor_approx.factor.variables
    }
    opt = OptFactor.from_approx(factor_approx, **kwargs)
    return opt.maximise(**p0)


maximize_factor_approx = maximise_factor_approx


def find_factor_mode(
        factor_approx: FactorApproximation,
        return_cov: bool = True,
        status: Status = Status(),
        min_iter: int = 2,
        opt_kws: Optional[dict] = None,
        **kwargs,
) -> OptResult:
    """ """
    opt_kws = {} if opt_kws is None else opt_kws

    opt = OptFactor.from_approx(factor_approx, **kwargs)
    res = opt.maximise(status=status, **opt_kws)

    if return_cov:
        # Calculate deterministic values
        value = factor_approx.factor(res.mode)
        res.mode.update(value.deterministic_values)

        # Calculate covariance of deterministic values
        jacobian = factor_approx.factor.jacobian(res.mode, opt.free_vars)
        update_det_cov(res, jacobian)

    return res


def laplace_factor_approx(
        model_approx: EPMeanField,
        factor: Factor,
        delta: float = 1.0,
        status: Status = Status(),
        opt_kws: Optional[Dict[str, Any]] = None,
):
    opt_kws = {} if opt_kws is None else opt_kws
    factor_approx = model_approx.factor_approximation(factor)
    res = find_factor_mode(factor_approx, return_cov=True, status=status, **opt_kws)

    model_dist = factor_approx.model_dist.from_mode_covariance(
        res.mode, res.hess_inv, res.log_norm
    )
    projection, status = factor_approx.project(
        model_dist, delta=delta, status=res.status
    )

    new_approx, status = model_approx.project(projection, status=status)

    return new_approx, status


class LaplaceOptimiser:
    def __init__(self, n_iter=4, delta=1.0, opt_kws: Optional[Dict[str, Any]] = None):
        self.history = dict()
        self.n_iter = n_iter
        self.delta = delta
        self.opt_kws = {} if opt_kws is None else opt_kws

    def step(
            self,
            model_approx,
            factors: Optional[List[Factor]] = None,
            status: Status = Status(),
    ) -> Iterator[Tuple[Factor, EPMeanField, Status]]:
        new_approx = model_approx
        factors = model_approx.factor_graph.factors if factors is None else factors
        for factor in factors:
            new_approx, status = laplace_factor_approx(
                new_approx, factor, self.delta, status=status, opt_kws=self.opt_kws
            )
            yield factor, new_approx, status

    def run(
            self,
            model_approx: EPMeanField,
            factors: Optional[List[Factor]] = None,
            status: Status = Status(),
    ) -> EPMeanField:
        new_approx = model_approx
        for i in range(self.n_iter):
            for factor, new_approx, status in self.step(new_approx, factors):
                self.history[i, factor] = new_approx
        return new_approx, status


class LeastSquaresOpt:
    _opt_params = dict(
        jac="2-point",
        method="trf",
        ftol=1e-08,
        xtol=1e-08,
        gtol=1e-08,
        x_scale=1.0,
        loss="linear",
        f_scale=1.0,
        diff_step=None,
        tr_solver=None,
        tr_options={},
        jac_sparsity=None,
        max_nfev=None,
        verbose=0,
    )

    def __init__(
            self,
            factor_approx: FactorApproximation,
            fixed_kws=None,
            param_bounds=None,
            opt_only=None,
            **kwargs,
    ):

        self.factor_approx = factor_approx
        self.opt_params = {**self._opt_params, **kwargs}

        param_shapes = {}
        param_bounds = {} if param_bounds is None else param_bounds
        fixed_kws = {} if fixed_kws is None else fixed_kws

        for v in factor_approx.factor.variables:
            dist = factor_approx.model_dist[v]
            if isinstance(dist, FixedMessage):
                fixed_kws[v] = dist.mean
            else:
                param_shapes[v] = dist.shape
                param_bounds[v] = dist._support

        self.fixed_kws = fixed_kws
        self.param_shapes = FlattenArrays(param_shapes)

        if opt_only is None:
            opt_only = tuple(
                v
                for v, d in factor_approx.cavity_dist.items()
                if not isinstance(d, FixedMessage)
            )

        self.opt_only = opt_only
        self.resid_means = {k: factor_approx.cavity_dist[k].mean for k in self.opt_only}
        self.resid_scales = {
            k: factor_approx.cavity_dist[k].scale for k in self.opt_only
        }
        self.resid_shapes = FlattenArrays(
            {k: np.shape(m) for k, m in self.resid_means.items()}
        )

        self.bounds = tuple(
            np.array(
                list(
                    zip(
                        *[
                            b
                            for k, s in param_shapes.items()
                            for bound in param_bounds[k]
                            for b in repeat(bound, np.prod(s, dtype=int))
                        ]
                    )
                )
            )
        )

    def __call__(self, arr):
        p0 = self.param_shapes.unflatten(arr)
        values = {**p0, **self.fixed_kws}
        fvals = self.factor_approx.factor(values)
        values.update(fvals.deterministic_values)
        residuals = {
            v: (values[v] - mean) / self.resid_scales[v]
            for v, mean in self.resid_means.items()
        }
        return self.resid_shapes.flatten(residuals)

    def least_squares(self, values=None):
        values = values or {}
        model_dist = self.factor_approx.model_dist
        p0 = {
            v: values[v] if v in values else model_dist[v].sample()
            for v in self.param_shapes.keys()
        }
        arr = self.param_shapes.flatten(p0)

        res = least_squares(self, arr, bounds=self.bounds, **self.opt_params)

        sol = self.param_shapes.unflatten(res.x)
        fval = self.factor_approx.factor({**sol, **self.fixed_kws})
        det_vars = fval.deterministic_values

        jac = {
            (d, k): b
            for k, a in self.param_shapes.unflatten(res.jac.T, ndim=1).items()
            for d, b in self.resid_shapes.unflatten(a.T, ndim=1).items()
        }
        hess = self.param_shapes.unflatten(res.jac.T.dot(res.jac))

        def inv(a):
            shape = np.shape(a)
            ndim = len(shape)
            if ndim:
                a = np.asanyarray(a)
                s = shape[: ndim // 2]
                n = np.prod(s, dtype=int)
                return np.linalg.inv(a.reshape(n, n)).reshape(s + s)
            else:
                return 1 / a

        invhess = {k: inv(h) for k, h in hess.items()}
        for det in det_vars:
            invhess[det] = 0.0
            for v in sol:
                invhess[det] += propagate_uncertainty(invhess[v], jac[det, v])

        mode = {**sol, **det_vars}
        return mode, invhess, res


def lstsq_laplace_factor_approx(
        model_approx: EPMeanField,
        factor: Factor,
        delta: float = 0.5,
        opt_kws: Optional[Dict[str, Any]] = None,
):
    """ """
    factor_approx = model_approx.factor_approximation(factor)

    opt = LeastSquaresOpt(factor_approx, **({} if opt_kws is None else opt_kws))

    mode, covar, result = opt.least_squares()
    message = (
        "optimise.lsq_sq_laplace_factor_approx: "
        f"nfev={result.nfev}, njev={result.njev}, "
        f"optimality={result.optimality}, "
        f"cost={result.cost}, "
        f"status={result.status}, message={result.message}",
    )
    status = Status(result.success, message)

    model_dist = MeanField(
        {v: factor_approx.factor_dist[v].from_mode(mode[v], covar.get(v)) for v in mode}
    )

    projection, status = factor_approx.project(model_dist, delta=delta, status=status)

    return model_approx.project(projection, status=status)



import numpy as np
import pytest

from autofit import graphical as graph
from autofit.messages.fixed import FixedMessage
from autofit.messages.normal import NormalMessage


def likelihood(z, y):
    expz = np.exp(-z)
    logp = -np.log1p(expz)
    log1p = -np.log1p(1 / expz)
    loglike = y * logp + (1 - y) * log1p
    return loglike.sum()


def likelihood_jacobian(z, y):
    expz = np.exp(-z)
    logp = -np.log1p(expz)
    log1p = -np.log1p(1 / expz)
    loglike = y * logp + (1 - y) * log1p

    jac_z = y - 1 / (1 + expz)
    jac_y = logp - log1p

    return loglike.sum(), (jac_z, jac_y)


@pytest.fixture(name="likelihood_factor")
def make_likelihood_factor(z_, y_, obs, dims):
    factor = graph.Factor(likelihood, z_, y_)
    return factor


@pytest.fixture(name="likelihood_factor_jac")
def make_likelihood_factor_jac(z_, y_, obs, dims):
    factor = graph.Factor(likelihood, z_, y_, factor_jacobian=likelihood_jacobian)
    return factor


@pytest.fixture(name="model")
def make_model(prior_a, prior_b, likelihood_factor_jac, linear_factor_jac):
    return likelihood_factor_jac * linear_factor_jac * prior_a * prior_b


@pytest.fixture(name="start_approx")
def make_start_approx(
        a_,
        b_,
        z_,
        x_,
        y_,
):
    a = np.array([[-1.3], [0.7]])
    b = np.array([-0.5])
    n_obs = 200
    n_features, n_dims = a.shape
    x = 2 * np.random.randn(n_obs, n_features)
    z = x.dot(a) + b
    p = 1 / (1 + np.exp(-z))
    y = np.random.binomial(1, p)

    return {
        a_: NormalMessage.from_mode(np.zeros((n_features, n_dims)), 10),
        b_: NormalMessage.from_mode(np.zeros(n_dims), 10),
        z_: NormalMessage.from_mode(np.zeros((n_obs, n_dims)), 100),
        x_: FixedMessage(x),
        y_: FixedMessage(y),
    }


def test_laplace(
        model,
        start_approx,
        y_,
        z_,
):
    model_approx = graph.EPMeanField.from_approx_dists(model, start_approx)
    laplace = graph.LaplaceOptimiser()
    opt = graph.EPOptimiser(model_approx.factor_graph, default_optimiser=laplace)
    new_approx = opt.run(model_approx, max_steps=10)

    y = new_approx.mean_field[y_].mean
    z_pred = new_approx(new_approx.mean_field.mean)[z_]
    y_pred = z_pred > 0
    (tpr, fpr), (fnr, tnr) = np.dot(
        np.array([y, 1 - y]).reshape(2, -1),
        np.array([y_pred, 1 - y_pred]).reshape(2, -1).T,
    )

    accuracy = (tpr + tnr) / (tpr + fpr + fnr + tnr)
    assert 0.95 > accuracy > 0.75



import os
import pickle
from pathlib import Path

import pytest

import autofit as af
from autoconf.conf import output_path_for_test
from autofit.non_linear.paths.null import NullPaths

from autofit.mock.mock import MockSearch

def test_null_paths():
    search = af.DynestyStatic()

    assert isinstance(
        search.paths,
        NullPaths
    )


class TestPathDecorator:
    @staticmethod
    def assert_paths_as_expected(paths):
        assert paths.name == "name"
        assert paths.path_prefix == ""

    def test_with_arguments(self):
        search = MockSearch()
        search.paths = af.DirectoryPaths(name="name")

        self.assert_paths_as_expected(search.paths)

    def test_positional(self):
        search = MockSearch("name")
        paths = search.paths

        assert paths.name == "name"

    def test_paths_argument(self):
        search = MockSearch()
        search.paths = af.DirectoryPaths(name="name")
        self.assert_paths_as_expected(search.paths)

    def test_combination_argument(self):
        search = MockSearch("other", )
        search.paths = af.DirectoryPaths(name="name")
        self.assert_paths_as_expected(search.paths)


output_path = Path(
    __file__
).parent / "path"


@pytest.fixture(
    name="model"
)
def make_model():
    return af.Model(
        af.Gaussian
    )


@output_path_for_test(
    output_path
)
def test_identifier_file(model):
    paths = af.DirectoryPaths()
    paths.model = model
    paths.search = af.DynestyStatic()
    paths.save_all({}, {}, [])

    assert os.path.exists(
        output_path / paths.identifier / ".identifier"
    )


def test_serialize(model):
    paths = af.DirectoryPaths()
    paths.model = model

    pickled_paths = pickle.loads(
        pickle.dumps(
            paths
        )
    )

    assert pickled_paths.model is not None



import numpy as np
import pytest

from autofit.graphical import (
    EPMeanField,
    LaplaceOptimiser,
    EPOptimiser,
)
from autofit.messages import FixedMessage, NormalMessage


@pytest.fixture(name="likelihood")
def make_likelihood(norm):
    def likelihood(z, y):
        return norm.logpdf(z - y).sum()

    return likelihood


@pytest.fixture(name="model")
def make_model(likelihood_factor, linear_factor, prior_a, prior_b):
    return likelihood_factor * linear_factor * prior_a * prior_b


@pytest.fixture(name="model_approx")
def make_model_approx(
        model,
        a_,
        b_,
        z_,
        x_,
        y_,
):
    a = np.array([[-1.3], [0.7]])
    b = np.array([-0.5])

    n_obs = 100
    n_features, n_dims = a.shape

    x = 5 * np.random.randn(n_obs, n_features)
    y = x.dot(a) + b + np.random.randn(n_obs, n_dims)

    message_a = NormalMessage.from_mode(np.zeros((n_features, n_dims)), 100)

    message_b = NormalMessage.from_mode(np.zeros(n_dims), 100)

    message_z = NormalMessage.from_mode(np.zeros((n_obs, n_dims)), 100)

    # return MeanFieldApproximation.from_kws(
    return EPMeanField.from_approx_dists(
        model,
        {
            a_: message_a,
            b_: message_b,
            z_: message_z,
            x_: FixedMessage(x),
            y_: FixedMessage(y),
        },
    )


def check_model_approx(model_approx, a_, b_, z_, x_, y_):
    mean_field = model_approx.mean_field
    x = model_approx.mean_field[x_].mean
    y = model_approx.mean_field[y_].mean

    X = np.c_[x, np.ones(len(x))]
    XTX = X.T.dot(X) + np.eye(3) / 10.0
    cov = np.linalg.inv(XTX)

    cov_a = cov[:2, :]
    cov_b = cov[2, :]

    # Analytic results
    mean_a = cov_a.dot(X.T.dot(y))
    mean_b = cov_b.dot(X.T.dot(y))
    a_std = cov_a.diagonal()[:, None] ** 0.5
    b_std = cov_b[[-1]] ** 0.5

    assert mean_field[a_].mean == pytest.approx(mean_a, rel=1e-2)
    assert mean_field[b_].mean == pytest.approx(mean_b, rel=1e-2)
    assert mean_field[a_].sigma == pytest.approx(a_std, rel=0.5)
    assert mean_field[b_].sigma == pytest.approx(b_std, rel=0.5)


@pytest.fixture(name="model_jac_approx")
def make_model_jac_approx(
        model, a_, b_, z_, x_, y_, likelihood_factor, linear_factor_jac, prior_a, prior_b
):
    a = np.array([[-1.3], [0.7]])
    b = np.array([-0.5])

    n_obs = 100
    n_features, n_dims = a.shape

    x = 5 * np.random.randn(n_obs, n_features)
    y = x.dot(a) + b + np.random.randn(n_obs, n_dims)

    # like = NormalMessage(y, np.ones_like(y)).as_factor(z_)
    model = likelihood_factor * linear_factor_jac * prior_a * prior_b

    model_jac_approx = EPMeanField.from_approx_dists(
        model,
        {
            a_: NormalMessage.from_mode(np.zeros((n_features, n_dims)), 100),
            b_: NormalMessage.from_mode(np.zeros(n_dims), 100),
            z_: NormalMessage.from_mode(np.zeros((n_obs, n_dims)), 100),
            x_: FixedMessage(x),
            y_: FixedMessage(y),
        },
    )
    return model_jac_approx


def test_jacobian(
        a_,
        b_,
        x_,
        z_,
        linear_factor,
        linear_factor_jac,
):
    n, m, d = 5, 4, 3
    x = np.random.rand(n, d)
    a = np.random.rand(d, m)
    b = np.random.rand(m)

    values = {x_: x, a_: a, b_: b}

    g0 = {z_: np.random.rand(n, m)}
    fval0, fjac0 = linear_factor.func_jacobian(values)
    fval1, fjac1 = linear_factor_jac.func_jacobian(values)

    fgrad0 = fjac0.grad(g0)
    fgrad1 = fjac1.grad(g0)

    assert fval0 == fval1
    assert (fgrad0 - fgrad1).norm() < 1e-6
    assert (fval0.deterministic_values - fval1.deterministic_values).norm() == 0


def test_laplace(
        model_approx,
        a_,
        b_,
        x_,
        y_,
        z_,
):
    laplace = LaplaceOptimiser()
    opt = EPOptimiser(model_approx.factor_graph, default_optimiser=laplace)
    model_approx = opt.run(model_approx)

    check_model_approx(model_approx, a_, b_, z_, x_, y_)


def test_laplace_jac(
        model_jac_approx,
        a_,
        b_,
        x_,
        y_,
        z_,
):
    laplace = LaplaceOptimiser()
    opt = EPOptimiser(model_jac_approx.factor_graph, default_optimiser=laplace)
    model_approx = opt.run(model_jac_approx)

    check_model_approx(model_approx, a_, b_, z_, x_, y_)



from os import path

import pytest

import autofit as af

from autofit.mock.mock_model import MockClassx3

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class TestLBFGSConfig:
    def test__loads_from_config_file_correct(self):
        lbfgs = af.LBFGS(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            tol=0.2,
            disp=True,
            maxcor=11,
            ftol=2.,
            gtol=3.,
            eps=4.,
            maxfun=25000,
            maxiter=26000,
            iprint=-2,
            maxls=21,
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
            iterations_per_update=10,
            number_of_cores=2,
        )

        assert lbfgs.prior_passer.sigma == 2.0
        assert lbfgs.prior_passer.use_errors is False
        assert lbfgs.prior_passer.use_widths is False
        assert lbfgs.config_dict_search["tol"] == 0.2
        assert lbfgs.config_dict_options["maxcor"] == 11
        assert lbfgs.config_dict_options["ftol"] == 2.
        assert lbfgs.config_dict_options["gtol"] == 3.
        assert lbfgs.config_dict_options["eps"] == 4.
        assert lbfgs.config_dict_options["maxfun"] == 25000
        assert lbfgs.config_dict_options["maxiter"] == 26000
        assert lbfgs.config_dict_options["iprint"] == -2
        assert lbfgs.config_dict_options["maxls"] == 21
        assert lbfgs.config_dict_options["disp"] == True
        assert isinstance(lbfgs.initializer, af.InitializerBall)
        assert lbfgs.initializer.lower_limit == 0.2
        assert lbfgs.initializer.upper_limit == 0.8
        assert lbfgs.iterations_per_update == 10
        assert lbfgs.number_of_cores == 2

        lbfgs = af.LBFGS()

        assert lbfgs.prior_passer.sigma == 3.0
        assert lbfgs.prior_passer.use_errors is True
        assert lbfgs.prior_passer.use_widths is True
        assert lbfgs.config_dict_search["tol"] == None
        assert lbfgs.config_dict_options["maxcor"] == 10
        assert lbfgs.config_dict_options["ftol"] == 2.220446049250313e-09
        assert lbfgs.config_dict_options["gtol"] == 1e-05
        assert lbfgs.config_dict_options["eps"] == 1e-08
        assert lbfgs.config_dict_options["maxfun"] == 15000
        assert lbfgs.config_dict_options["maxiter"] == 15000
        assert lbfgs.config_dict_options["iprint"] == -1
        assert lbfgs.config_dict_options["maxls"] == 20
        assert lbfgs.config_dict_options["maxiter"] == 15000
        assert lbfgs.config_dict_options["disp"] == False
        assert isinstance(lbfgs.initializer, af.InitializerPrior)
        assert lbfgs.iterations_per_update == 11

    def test__samples_from_model(self):
        
        lbfgs = af.LBFGS()
        lbfgs.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "LBFGS"))
        lbfgs.paths._identifier = "tag"

        model = af.ModelMapper(mock_class=MockClassx3)
        model.mock_class.one = af.LogUniformPrior(lower_limit=1e-8, upper_limit=100.0)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=100.0)
        model.mock_class.three = af.LogUniformPrior(lower_limit=1e-8, upper_limit=100.0)

        samples = lbfgs.samples_from(model=model)

        assert isinstance(samples.parameter_lists, list)
        assert isinstance(samples.parameter_lists[0], list)
        assert isinstance(samples.log_likelihood_list, list)
        assert isinstance(samples.log_prior_list, list)
        assert isinstance(samples.log_posterior_list, list)

        assert samples.parameter_lists[0] == pytest.approx(
            [50.005469, 25.143677, 10.06950], 1.0e-4
        )

        assert samples.log_likelihood_list[0] == pytest.approx(-45.134121, 1.0e-4)
        assert samples.log_posterior_list[0] == pytest.approx(-44.97504284, 1.0e-4)
        assert samples.weight_list[0] == 1.0

        assert len(samples.parameter_lists) == 1
        assert len(samples.log_likelihood_list) == 1



import pytest

import autofit as af
from autofit import graphical as g
from autofit.mock.mock import MockAnalysis


@pytest.fixture(
    name="prior"
)
def make_prior(model_gaussian_x1):
    return model_gaussian_x1.priors[0]


def test_collection_prior_model(
        prior
):
    assert af.Collection(prior).prior_count == 1


@pytest.fixture(
    name="prior_factor"
)
def make_prior_factor(prior):
    return g.PriorFactor(
        prior
    )


def test_prior_factor(
        prior_factor
):
    assert prior_factor.prior_model.prior_count == 1


def test_log_likelihood_function(
        prior_factor
):
    instance = prior_factor.prior_model.instance_from_prior_medians()
    assert prior_factor.log_likelihood_function(
        instance
    ) == pytest.approx(-0.9189385332046727, 1.0e-8)


def test_optimise(model_gaussian_x1, prior):
    optimizer = af.DynestyStatic(
        maxcall=10
    )
    analysis = MockAnalysis()
    factor = g.AnalysisFactor(
        model_gaussian_x1,
        analysis
    )
    prior_factor = factor.prior_factors[0]
    result, status = optimizer.optimise(
        prior_factor,
        factor.mean_field_approximation()
    )

    assert status

    optimized_mean = list(result.mean_field.values())[0].mean
    assert optimized_mean == pytest.approx(
        prior.mean,
        rel=0.1
    )



from itertools import combinations

import numpy as np
import pytest

try:
    import jax

    _HAS_JAX = True
except ImportError:
    _HAS_JAX = False

from autofit.mapper.variable import variables
from autofit.graphical.factor_graphs import (
    Factor,
    FactorValue,
)


def test_jacobian_equiv():
    if not _HAS_JAX:
        return

    def linear(x, a, b, c):
        z = x.dot(a) + b
        return (z ** 2).sum(), z

    x_, a_, b_, c_, z_ = variables("x, a, b, c, z")
    x = np.arange(10.0).reshape(5, 2)
    a = np.arange(2.0).reshape(2, 1)
    b = np.ones(1)
    c = -1.0

    factors = [
        Factor(
            linear,
            x_,
            a_,
            b_,
            c_,
            factor_out=(FactorValue, z_),
            numerical_jacobian=False,
        ),
        Factor(
            linear,
            x_,
            a_,
            b_,
            c_,
            factor_out=(FactorValue, z_),
            numerical_jacobian=False,
            jacfwd=False,
        ),
        Factor(
            linear,
            x_,
            a_,
            b_,
            c_,
            factor_out=(FactorValue, z_),
            numerical_jacobian=False,
            vjp=True,
        ),
        Factor(
            linear,
            x_,
            a_,
            b_,
            c_,
            factor_out=(FactorValue, z_),
            numerical_jacobian=True,
        ),
    ]

    values = {x_: x, a_: a, b_: b, c_: c}
    outputs = [factor.func_jacobian(values) for factor in factors]

    tol = pytest.approx(0, abs=1e-4)
    pairs = combinations(outputs, 2)
    g0 = FactorValue(1.0, {z_: np.ones((5, 1))})
    for (val1, jac1), (val2, jac2) in pairs:
        assert val1 == val2

        # test with different ways of calculating gradients
        grad1, grad2 = jac1.grad(g0), jac2.grad(g0)
        assert (grad1 - grad2).norm() == tol
        grad1 = g0.to_dict() * jac1
        assert (grad1 - grad2).norm() == tol
        grad2 = g0.to_dict() * jac2
        assert (grad1 - grad2).norm() == tol

        grad1, grad2 = jac1.grad(val1), jac2.grad(val2)
        assert (grad1 - grad2).norm() == tol

        # test getting gradient with no args
        assert (jac1.grad() - jac2.grad()).norm() == tol


def test_jac_model():
    if not _HAS_JAX:
        return

    def linear(x, a, b):
        z = x.dot(a) + b
        return (z ** 2).sum(), z

    def likelihood(y, z):
        return ((y - z) ** 2).sum()

    def combined(x, y, a, b):
        like, z = linear(x, a, b)
        return like + likelihood(y, z)

    x_, a_, b_, y_, z_ = variables("x, a, b, y, z")
    x = np.arange(10.0).reshape(5, 2)
    a = np.arange(2.0).reshape(2, 1)
    b = np.ones(1)
    y = np.arange(0.0, 10.0, 2).reshape(5, 1)
    values = {x_: x, y_: y, a_: a, b_: b}
    linear_factor = Factor(linear, x_, a_, b_, factor_out=(FactorValue, z_), vjp=True)
    like_factor = Factor(likelihood, y_, z_, vjp=True)
    full_factor = Factor(combined, x_, y_, a_, b_, vjp=True)
    model_factor = like_factor * linear_factor

    x = np.arange(10.0).reshape(5, 2)
    a = np.arange(2.0).reshape(2, 1)
    b = np.ones(1)
    y = np.arange(0.0, 10.0, 2).reshape(5, 1)
    values = {x_: x, y_: y, a_: a, b_: b}

    # Fully working problem
    fval, jac = full_factor.func_jacobian(values)
    grad = jac.grad()

    model_val, model_jac = model_factor.func_jacobian(values)
    model_grad = model_jac.grad()

    linear_val, linear_jac = linear_factor.func_jacobian(values)
    like_val, like_jac = like_factor.func_jacobian(
        {**values, **linear_val.deterministic_values}
    )
    combined_val = like_val + linear_val

    # Manually back propagate
    combined_grads = linear_jac.grad(like_jac.grad())

    vals = (fval, model_val, combined_val)
    grads = (grad, model_grad, combined_grads)
    pairs = combinations(zip(vals, grads), 2)
    for (val1, grad1), (val2, grad2) in pairs:
        assert val1 == val2
        assert (grad1 - grad2).norm() == pytest.approx(0, 1e-6)



import pytest

import autofit as af
from autoconf.exc import ConfigException
from autofit.mapper.model_object import Identifier
from autofit.mock.mock_model import MockWithTuple


class SomeWeirdClass:
    def __init__(self, argument):
        self.argument = argument


def test_config_error():
    model = af.Model(
        SomeWeirdClass
    )

    with pytest.raises(ConfigException):
        print(Identifier([
            model
        ]))


def test_mapper_from_prior_arguments_simple_collection():
    old = af.UniformPrior()
    new = af.UniformPrior()
    collection = af.Collection(
        value=old
    )
    collection = collection.mapper_from_prior_arguments({
        old: new
    })

    assert collection.value == new


def test_direct_instances_only():
    child = af.Model(
        af.Gaussian,
        centre=0.0,
        normalization=0.1,
        sigma=0.01,
    )
    child.constant = 1.0

    model = af.Model(
        af.Gaussian,
        centre=child,
        normalization=0.1,
        sigma=0.01,
    )

    new_model = model.gaussian_prior_model_for_arguments({})
    assert not hasattr(new_model, "constant")


def test_function_from_instance():
    assert af.PriorModel.from_instance(
        test_function_from_instance
    ) is test_function_from_instance


def test_as_model_tuples():
    model = af.Model(MockWithTuple)
    assert isinstance(
        model.tup.tup_0,
        af.UniformPrior
    )
    assert isinstance(
        model.tup.tup_1,
        af.UniformPrior
    )

    instance = model.instance_from_prior_medians()
    assert instance.tup == (0.5, 0.5)

    model = af.AbstractPriorModel.from_instance(
        instance
    )
    assert model.tup == (0.5, 0.5)
    assert model.info == """tup                                                                                       (0.5, 0.5)"""


def test_set_centre():
    model = af.Model(MockWithTuple)
    model.tup_0 = 10.0

    instance = model.instance_from_prior_medians()
    assert instance.tup[0] == 10.0

    model = af.Model(MockWithTuple)
    model.tup.tup_0 = 10.0

    instance = model.instance_from_prior_medians()
    assert instance.tup[0] == 10.0


def test_passing_priors():
    model = af.Model(
        MockWithTuple
    )

    new_model = model.mapper_from_gaussian_tuples([
        (1, 1),
        (1, 1),
    ])
    assert isinstance(new_model.tup_0, af.GaussianPrior)
    assert isinstance(new_model.tup_1, af.GaussianPrior)


def test_passing_fixed():
    model = af.Model(
        MockWithTuple
    )
    model.tup_0 = 0.1
    model.tup_1 = 2.0

    new_model = model.mapper_from_gaussian_tuples([])
    assert new_model.tup_0 == 0.1
    assert new_model.tup_1 == 2.0



import autofit as af
from os import path
import os
import matplotlib.pyplot as plt
import numpy as np


class Analysis(af.Analysis):
    def __init__(self, data, noise_map):

        super().__init__()

        self.data = data
        self.noise_map = noise_map

    def log_likelihood_function(self, instance):
        """
        Determine the log likelihood of a fit of multiple profiles to the dataset.

        Parameters
        ----------
        instance : af.Collection
            The model instances of the profiles.

        Returnsn
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the dataset.

        The `instance` that comes into this method is a Collection. It contains instances of every class
        we instantiated it with, where each instance is named following the names given to the Collection,
        which in this example is a `Gaussian` (with name `gaussian) and Exponential (with name `exponential`):
        """

        xvalues = np.arange(self.data.shape[0])

        try:
            model_data = sum(
                line.profile_1d_via_xvalues_from(xvalues=xvalues) for line in instance
            )
        except TypeError:
            model_data = instance.profile_1d_via_xvalues_from(xvalues=xvalues)

        residual_map = self.data - model_data
        chi_squared_map = (residual_map / self.noise_map) ** 2.0
        log_likelihood = -0.5 * sum(chi_squared_map)

        return log_likelihood

    def visualize(self, paths, instance, during_analysis):

        """
        During a model-fit, the `visualize` method is called throughout the non-linear search. The `instance` passed
        into the visualize method is maximum log likelihood solution obtained by the model-fit so far and it can be
        used to provide on-the-fly images showing how the model-fit is going.
        """

        xvalues = np.arange(self.data.shape[0])

        try:
            model_data = sum(
                line.profile_1d_via_xvalues_from(xvalues=xvalues) for line in instance
            )
        except TypeError:
            model_data = instance.profile_1d_via_xvalues_from(xvalues=xvalues)

        plt.errorbar(
            x=xvalues,
            y=self.data,
            yerr=self.noise_map,
            color="k",
            ecolor="k",
            elinewidth=1,
            capsize=2,
        )
        plt.plot(range(self.data.shape[0]), model_data, color="r")
        plt.title("Dynesty model fit to 1D Gaussian + Exponential dataset.")
        plt.xlabel("x values of profile")
        plt.ylabel("Profile normalization")

        os.makedirs(paths.image_path, exist_ok=True)
        plt.savefig(path.join(paths.image_path, "model_fit.png"))
        plt.clf()
